import os
from collections.abc import Sequence
from typing import Any, Mapping

import mutalyzer_hgvs_parser
from flask import Flask
from jinja2 import Environment, FileSystemLoader

from gtgt.mutalyzer import init_description
from gtgt.transcript import Result

from . import Transcript
from .variant_validator import lookup_variant

hgvs_error = (
    mutalyzer_hgvs_parser.exceptions.UnexpectedCharacter,
    mutalyzer_hgvs_parser.exceptions.UnexpectedEnd,
)

app = Flask(__name__)


payload = dict[str, Any] | None


def render(
    template_file: str,
    variant: str | None = None,
    results: Sequence[Result] = [],
    links: Mapping[str, str] | None = None,
    error: Mapping[str, str] | None = None,
) -> str:
    """Render the html template using jinja2 directly"""
    root = os.path.dirname(__file__)
    template_file = os.path.join(root, template_file)
    template_folder = os.path.dirname(template_file)
    environment = Environment(loader=FileSystemLoader(template_folder))
    template = environment.get_template(os.path.basename(template_file))

    # Exclude any payloads that are not set
    d: dict[str, Any] = dict()
    if variant:
        d["variant"] = variant
    if results:
        d["results"] = organize_results(results)
    if links:
        d["links"] = links
    if error:
        d["error"] = error

    return template.render(**d)


def organize_results(
    results: Sequence[Result],
) -> dict[str, list[Result] | Result]:
    """
    The Results are grouped in a dict with the following keys
    - input, which contains the Result for the input
    - modified, which contains the Results where one or more of the input
      varianst have been changed
    - all, which contains all other Results, including the wildtype etc, in
      the original order
    """

    # Keep track of the Results we still have to precess
    todo = [x for x in results]

    # Extract the input and wildtype results, wich are special
    input_ = next((x for x in results if x.therapy.name == "Input"))
    wildtype = next((x for x in results if x.therapy.name == "Wildtype"))
    todo.remove(input_)
    todo.remove(wildtype)

    # Extract the variants from the input
    input_variants = input_.therapy.variants

    # Find all Results where one or more of the input variants are modified or removed
    modified = list()
    for result in todo:
        for variant in input_variants:
            if variant not in result.therapy.variants:
                modified.append(result)
                break

    # Exclude the Results that modify the input variant
    todo = [x for x in todo if x not in modified]

    return {
        "input": [input_, wildtype],
        "modified": modified,
        "all": todo,
    }


def validate_user_input(input: str) -> Mapping[str, str]:
    """
    Validate the user input

    If there is an error, return a dict with summary and details of the error
    """
    error = dict()

    # Test if the variant is valid HGVS
    try:
        mutalyzer_hgvs_parser.to_model(input)
    except hgvs_error as e:
        error["summary"] = "Not a valid HGVS description"
        error["details"] = str(e)
        return error

    return error


@app.route("/")
@app.route("/<variant>")
def result(variant: str | None = None) -> str:
    template_file = "templates/index.html.j2"

    # If no variant was specified
    if not variant:
        return render(template_file)

    # Invalid user input
    if error := validate_user_input(variant):
        return render(template_file, variant=variant, error=error)

    # Analyze the transcript
    try:
        transcript = Transcript.from_description(init_description(variant))
        results = transcript.analyze(variant)
    except Exception as e:
        error = {"summary": "Analysis failed", "details": str(e)}
        results = []

    # Get external links
    try:
        links = lookup_variant(variant).url_dict()
    except Exception as e:
        links = dict()

    if error:
        return render(template_file, variant=variant, error=error)
    else:
        return render(template_file, results=results, links=links, variant=variant)



from collections.abc import Sequence

import svg
from exonviz.draw import draw_exons
from exonviz.mutalyzer import build_exons
from mutalyzer.description import Description

config = {
    "width": 768,
    "height": 20,
    "scale": 1.0,
    "noncoding": True,
    "gap": 0,
    "color": "#4C72B7",
    "exonnumber": True,
    "firstexon": 1,
    "lastexon": 9999,
    "variantcolors": ["#BA1C30", "#DB6917", "#EBCE2B", "#702C8C", "#C0BD7F"],
    "variantshape": "pin",
}


def draw(d: Description) -> str:
    exons, dropped_variants = build_exons(
        hgvs=d.input_description,
        mutalyzer=d.output()["selector_short"],
        config=config,
    )

    # Determine the minimum scale at which we can draw every exon
    min_scale = max([e.min_scale() for e in exons])
    config["scale"] = max(min_scale, 1)

    fig = draw_exons(exons, config=config)

    # Make the figure scalable with CSS by resetting the width and height
    # and setting a viewBox
    width = fig.width
    height = fig.height
    fig.viewBox = svg.ViewBoxSpec(0, 0, width, height)

    fig.width = None
    fig.height = None

    return str(fig)



import dataclasses
import logging
from copy import deepcopy
from typing import Any, Mapping, Sequence

from mutalyzer.description import Description

from .bed import Bed
from .exonviz import draw
from .mutalyzer import (
    Therapy,
    Variant,
    generate_therapies,
    get_chrom_name,
    get_exons,
    get_offset,
    get_strand,
    init_description,
    mutation_to_cds_effect,
    protein_prediction,
    sequence_from_description,
)

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Comparison:
    name: str
    percentage: float
    basepairs: str

    @classmethod
    def from_dict(cls, dict: Mapping[str, Any]) -> "Comparison":
        """Create a Variant object from a dict representation of a Variant"""
        return cls(**dict)


@dataclasses.dataclass
class Result:
    """To hold results for separate mutations of a transcript"""

    therapy: Therapy
    comparison: Sequence[Comparison]

    def __gt__(self, other: "Result") -> bool:
        """Sort Result based on the sum of the percentage"""
        if not isinstance(other, Result):
            msg = f"Unsupported comparison between Bed and {type(other)}"
            raise NotImplementedError(msg)

        total_self = sum(c.percentage for c in self.comparison)
        total_other = sum(c.percentage for c in other.comparison)
        return total_self > total_other

    @classmethod
    def from_dict(cls, dict: Mapping[str, Any]) -> "Result":
        return cls(
            therapy=Therapy.from_dict(dict["therapy"]),
            comparison=[Comparison.from_dict(x) for x in dict["comparison"]],
        )


class Transcript:
    def __init__(self, rna_features: list[Bed], protein_features: list[Bed]):
        self.rna_features = rna_features
        self.protein_features = protein_features

    @classmethod
    def from_description(cls, d: Description) -> "Transcript":
        """Create a Transcript object from a mutalyzer Description"""
        # Check if we can use this Description to initialize a Transcript
        offset = get_offset(d)

        # Get exons and add the offset
        selector_model = d.get_selector_model()
        exons = selector_model["exon"]
        exons = [(start + offset, end + offset) for start, end in exons]

        # Get CDS and add the offset
        cds = selector_model["cds"][0]
        cds = (cds[0] + offset, cds[1] + offset)

        # Get the strand and chromosome name
        chrom = get_chrom_name(d)
        strand = get_strand(d)

        # Convert to Bed to intersect the exons with the CDS to get the coding
        # exons
        exon_bed = Bed.from_blocks(chrom, exons)
        exon_bed.name = "Exons"
        exon_bed.strand = strand

        cds_bed = Bed.from_blocks(chrom, [cds])
        cds_bed.strand = strand

        # Make a copy to intersect
        coding_exons = exon_bed.intersect(cds_bed)
        coding_exons.name = "Coding exons"

        return cls(rna_features=[exon_bed], protein_features=[coding_exons])

    def records(self) -> Sequence[Bed]:
        """Return the Bed records that make up the Transcript"""
        return self.rna_features + self.protein_features

    def rna_records(self) -> list[Bed]:
        """Return the Bed records that contain RNA features"""
        return self.rna_features

    def protein_records(self) -> list[Bed]:
        """Return the Bed records that contain protein features"""
        return self.protein_features

    def intersect(self, selector: Bed) -> None:
        """Update transcript to only contain features that intersect the selector"""
        for record in self.records():
            record.intersect(selector)

    def overlap(self, selector: Bed) -> None:
        """Update transcript to only contain features that overlap the selector"""
        for record in self.records():
            record.overlap(selector)

    def subtract(self, selector: Bed) -> None:
        """Remove all features from transcript that intersect the selector"""
        for record in self.records():
            record.subtract(selector)

    def compare(self, other: object) -> Sequence[Comparison]:
        """Compare the size of each record in the transcripts"""
        if not isinstance(other, Transcript):
            raise NotImplementedError

        # Compare each record that makes up self and other
        # The comparison will fail if the record.name does not match
        cmp = list()
        for record1, record2 in zip(self.records(), other.records()):
            percentage = record1.compare(record2)
            fraction = record1.compare_basepair(record2)
            C = Comparison(record1.name, percentage, fraction)
            cmp.append(C)

        return cmp

    def compare_score(self, other: object) -> float:
        """Compare the size of each records in the transcripts

        Returns the average value for all records
        """
        if not isinstance(other, Transcript):
            raise NotImplementedError
        cmp = self.compare(other)

        values = [x.percentage for x in cmp]
        return sum(values) / len(cmp)

    def mutate(self, d: Description, variants: Sequence[Variant]) -> None:
        """Mutate the transcript based on the specified variants"""
        # Determine the chromosome the transcript is on
        if self.records():
            chrom = self.records()[0].chrom
        else:
            chrom = ""

        # Update protein features
        protein_changes = Bed.from_blocks(chrom, mutation_to_cds_effect(d, variants))
        for record in self.protein_records():
            record.subtract(protein_changes)

        # Update RNA features
        rna_changes = Bed.from_blocks(
            chrom, [v.genomic_coordinates(d) for v in variants]
        )
        for record in self.rna_records():
            self.subtract(rna_changes)

    def analyze(self, hgvs: str) -> Sequence[Result]:
        """Analyze the transcript based on the specified HGVS description"""

        # r. notations are not supported
        coordinate_system = hgvs.split(":")[1][0:1]
        if coordinate_system != "c":
            raise NotImplementedError(
                f"Coordinate system '{coordinate_system}' is not supported"
            )
        # Initialize the input HGVS description
        d = init_description(hgvs)

        # Extract the input variants as internal delins
        sequence = sequence_from_description(d)
        input_variants = [
            Variant.from_model(delins, sequence=sequence)
            for delins in d.delins_model["variants"]
        ]

        results = list()

        # Store the wildtype
        wt = Therapy(
            name="Wildtype",
            hgvsc=hgvs.split("c.")[0] + "c.=",
            hgvsr=hgvs.split("c.")[0] + "r.=",
            hgvsp=protein_prediction(d, [])[0],
            description="These are the annotations as defined on the reference. They are always 100% by definition.",
            variants=list(),
        )
        wildtype = Result(wt, self.compare(self))
        results.append(wildtype)

        # Store the input variants as Therapy
        input = Therapy(
            name="Input",
            hgvsc=hgvs,
            hgvsr=hgvs.replace(":c.", ":r."),
            hgvsp=protein_prediction(d, input_variants)[0],
            description="The annotations based on the supplied input variants.",
            variants=input_variants,
            figure=draw(d),
        )
        patient = deepcopy(self)
        patient.mutate(d, input.variants)
        results.append(Result(input, patient.compare(self)))

        # Generate all possible therapies
        for therapy in generate_therapies(d):

            # Apply the combination to the wildtype transcript
            modified_transcript = deepcopy(self)
            modified_transcript.mutate(d, therapy.variants)
            results.append(Result(therapy, modified_transcript.compare(self)))

        # Order the results
        wt_patient = results[:2]
        rest = sorted(results[2:], reverse=True)
        return wt_patient + rest

    def __str__(self) -> str:
        return "\n".join(str(record) for record in self.records())

    @property
    def exons(self) -> Bed | None:
        """Helper function to get the exons for Transcript

        Note that this relies on the Bed record for the exons to be
        named 'Exons'
        """
        for record in self.rna_features:
            if record.name == "Exons":
                return record
        else:
            return None

    @property
    def coding_exons(self) -> Bed | None:
        """Helper function to get the coding exons for Transcript

        Note that this relies on the Bed record for the coding exons to be
        named 'Coding exons'
        """
        for record in self.protein_features:
            if record.name == "Coding exons":
                return record
        else:
            return None



import logging
from collections import defaultdict, namedtuple
from itertools import combinations
from typing import Any, Mapping, Sequence

from mutalyzer.description import Description

from gtgt.bed import Bed
from gtgt.mutalyzer import (
    get_assembly_name,
    get_chrom_name,
    get_offset,
    get_transcript_name,
)
from gtgt.range import Range, overlap
from gtgt.variant_validator import parse_payload

from .models import Assembly, EnsemblTranscript
from .provider import UCSC, MyGene, Provider, VariantValidator, payload

logger = logging.getLogger(__name__)

ENSEMBL_TO_UCSC = {
    Assembly.HUMAN: "hg38",
    Assembly.RAT: "rn6",
}

# Supported tracks which contain protein features
PROTEIN_TRACKS: list[str] = [
    "unipDomain",
    "unipRepeat",
    "unipStruct",
    "unipOther",
    "unipMut",
    "unipModif",
    "unipDisulfBond",
    "unipChain",  # Unclear naming, in DMD the track name is just '1'
    "unipLocCytopl",
    "unipLocTransMemb",
    "unipLocExtra",
    "unipLocSignal",
    "unipInterest",
]
# Supported tracks which contain RNA features
RNA_TRACKS: list[str] = ["knownGene", "clinvarMain"]


# Holder for the required parameters for UCSC query
Parameters = namedtuple("Parameters", ["genome", "chrom", "start", "end", "track"])


def _lookup_track_payload(
    d: Description, track: str, ucsc: Provider = UCSC()
) -> payload:
    """Lookup the track payload for the specified Description"""
    # Determine the assembly
    if get_assembly_name(d) != "GRCh38":
        raise ValueError(f"Assembly {get_assembly_name(d)} for not supported ({d})")
    else:
        # Assembly as used in UCSC
        genome = "hg38"

    # Determine the start and end of the transcript of interest
    offset = get_offset(d)
    exons = d.get_selector_model()["exon"]
    transcript_start = exons[0][0] + offset
    transcript_end = exons[-1][1] + offset

    # Next, determine the uniprot ID for the protein domain
    parameters = Parameters(
        genome=genome,
        chrom=get_chrom_name(d),
        start=transcript_start,
        end=transcript_end,
        track=track,
    )
    return ucsc.get(parameters)


def _track_to_range(track: payload) -> list[Range]:
    """Extract a Range from a track"""
    blocks = list()
    chrom_start = track["chromStart"]
    chrom_end = track["chromEnd"]

    # If the track only contains a single block
    if track.get("blockCount", 1) == 1:
        blocks.append((chrom_start, chrom_end))
    else:
        starts = [int(x) for x in track["chromStarts"].strip(",").split(",")]
        sizes = [int(x) for x in track["blockSizes"].strip(",").split(",")]
        for start, size in zip(starts, sizes):
            block_start = chrom_start + start
            block_end = chrom_start + start + size
            blocks.append((block_start, block_end))
    return blocks


def _blocks_overlap(blocks: list[Range]) -> bool:
    """Determine if blocks overlap"""
    for a, b in combinations(blocks, 2):
        if overlap(a, b):
            return True
    else:
        return False


def _track_name(track: payload) -> str:
    """Determine the name of a track

    The format is a bit weird, the most informative name is in the 'comments'
    field, while 'name' appears to be truncated sometimes
    """
    name = track.get("name")
    comments = track.get("comments")

    if not name:
        raise ValueError(f"Name is missing from track {track}")

    if not comments:
        return str(name)
    else:
        return f"{name}: {comments}"


def _tracks_to_bed(tracks: Sequence[payload]) -> list[Bed]:
    """Convert a list of json track objects from UCSC into a list of Bed records

    Note that trakcs with the same name will be merged into a single Bed record
    """
    # Store the resulting Bed records
    records: list[Bed] = list()

    # Next, group the tracks by name
    grouped = defaultdict(list)
    for track in tracks:
        name = _track_name(track)
        grouped[name].append(track)

    for tracks in grouped.values():
        if not tracks:
            raise RuntimeError(tracks)
        # Look in the first record for all constant values
        track = tracks[0]
        # Get the blocks
        blocks: list[tuple[int, int]] = list()
        for track in tracks:
            blocks += _track_to_range(track)

        # Get the chromosome
        chrom = track["chrom"]
        name = _track_name(track)

        # Ensure that the blocks do not overlap
        if _blocks_overlap(blocks):
            raise ValueError(f"Found overlapping blocks in {name}")

        bed = Bed.from_blocks(chrom, blocks)
        # The color is stored in the 'reserved' field by UCSC
        bed.itemRgb = tuple(map(int, track["reserved"].split(",")))
        bed.score = track["score"]
        bed.strand = track["strand"]
        bed.name = name

        records.append(bed)

    return records


def lookup_track(
    d: Description,
    track: str,
    ucsc: Provider = UCSC(),
    mygene: Provider = MyGene(),
    variantvalidator: Provider = VariantValidator(),
) -> list[Bed]:
    """Lookup the specified track on UCSC"""
    # Required to match protein features to the correct transcript
    uniprot_id = None

    # For protein tracks, we have to lookup the uniprot_id
    if track in PROTEIN_TRACKS:
        # First, we look up the ENSG using VariantValidator
        variant = str(d)
        vv = variantvalidator.get(("hg38", variant))
        ensg = parse_payload(vv, variant, "hg38")["ensembl_gene_id"]
        assert ensg is not None

        # Look up protein ID to match the correct transcript
        uniprot_id = mygene.get((ensg,))["uniprot"]["Swiss-Prot"]
    elif track in RNA_TRACKS:
        pass
    else:
        raise ValueError(f"Track {track} is not supported, please ask to have it added")

    payload = _lookup_track_payload(d, track, ucsc)
    # Get only the tracks that match the specified uniprot ID
    if uniprot_id:
        tracks = [
            track for track in payload[track] if track.get("uniProtId") == uniprot_id
        ]
    else:
        tracks = [track for track in payload[track]]
    return _tracks_to_bed(tracks)


def chrom_to_uscs(seq_region_name: str) -> str:
    return "chrM" if seq_region_name == "MT" else f"chr{seq_region_name}"


def lookup_knownGene(
    transcript: EnsemblTranscript, track_name: str
) -> Mapping[str, Any]:

    provider = UCSC()

    parameters = Parameters(
        genome="hg38",
        chrom=chrom_to_uscs(transcript.seq_region_name),
        start=transcript.start,
        end=transcript.end,
        track=track_name,
    )
    track = provider.get(parameters)
    ts = f"{transcript.id}.{transcript.version}"
    track[track_name] = [
        entry for entry in track[track_name] if entry.get("name") == ts
    ]
    return track



import gzip
import json
import logging
import os
import urllib.request
from abc import ABC, abstractmethod
from typing import Any, Protocol
from urllib.error import HTTPError

payload = dict[str, Any]
parameters = tuple[Any, ...]

logger = logging.getLogger(__name__)


class Provider(Protocol):
    """Protocol for the Provider class

    There is also an abstrac base class _Provider, which implements file system
    caching of API calls and can be inherited from
    """

    def get(self, parameters: parameters) -> payload: ...


class _Provider(ABC):
    def __init__(self) -> None:
        cache = os.environ.get("GTGT_CACHE")
        name = type(self).__name__

        self.cache: str | None
        # Put the cache for each Provider in a separate folder
        if cache:
            self.cache = f"{cache}/{name}"
        else:
            self.cache = None

        # Ensure the cache folder exists
        if self.cache:
            os.makedirs(self.cache, exist_ok=True)

    @abstractmethod
    def get(self, parameters: parameters) -> payload:
        pass

    def __str__(self) -> str:
        return f"{type(self).__name__}(cache={self.cache})"

    def _fetch_url(self, url: str) -> payload:
        logger.info(f"Fetching {url=}")
        try:
            response = urllib.request.urlopen(url)
        except HTTPError as e:
            raise RuntimeError(str(e))

        data = response.read()

        try:
            js: payload = json.loads(data)
        except Exception as e:
            logger.error(data)
            raise e

        return js

    def _fname(self, parameters: parameters) -> str:
        """Generate file name for the cache based on the parameters"""
        return f"{self.cache}/{'_'.join(str(x) for x in parameters)}.json.gz"

    def _get(self, url: str, parameters: parameters) -> payload:
        """Get the requested data, from the filename or the url"""
        # If the cache is not enabled
        if not self.cache:
            return self._fetch_url(url)

        # Filename for the cached payload
        fname = self._fname(parameters)

        js: payload = dict()
        # If the payload is already in the cache
        if os.path.exists(fname):
            logger.info(f"Reading payload from {fname}")
            with gzip.open(fname, "rt") as fin:
                js = json.load(fin)
        else:
            # If the payload is not in the cache
            js = self._fetch_url(url)
            with gzip.open(fname, "wt") as fout:
                print(json.dumps(js), file=fout)
        return js


class MyGene(_Provider):
    def get(self, parameters: parameters) -> payload:
        ensembl_gene_id, *rest = parameters
        url = f"https://mygene.info/v3/gene/{ensembl_gene_id}?fields=uniprot"

        return self._get(url, parameters)


class VariantValidator(_Provider):
    def get(self, parameters: parameters) -> payload:
        prefix = "https://rest.variantvalidator.org/VariantValidator"
        suffix = "mane_select?content-type=application/json"

        assembly, variant, *rest = parameters
        if variant.startswith("ENS"):
            url = f"{prefix}/variantvalidator_ensembl/{assembly}/{variant}/{suffix}"
        else:
            url = f"{prefix}/variantvalidator/{assembly}/{variant}/{suffix}"

        return self._get(url, parameters)


class UCSC(_Provider):
    def get(self, parameters: parameters) -> payload:
        genome, chrom, start, end, track = parameters
        url = ";".join(
            (
                f"https://api.genome.ucsc.edu/getData/track?genome={genome}",
                f"chrom={chrom}",
                f"start={start}",
                f"end={end}",
                f"track={track}",
            )
        )
        return self._get(url, parameters)


class Ensembl(_Provider):
    def get(self, parameters: parameters) -> payload:
        transcript, *rest = parameters
        url = f"http://rest.ensembl.org/lookup/id/{transcript}?content-type=application/json"
        return self._get(url, parameters)



import dataclasses
import logging
from copy import deepcopy
from typing import Any, Mapping, Sequence, TypeVar

import Levenshtein
import mutalyzer_hgvs_parser
from mutalyzer.converter.to_hgvs_coordinates import to_hgvs_locations
from mutalyzer.converter.variants_de_to_hgvs import (
    delins_to_del,
    delins_to_delins,
    delins_to_duplication,
    delins_to_insertion,
    delins_to_repeat,
    delins_to_substitution,
    get_end,
    get_start,
    is_duplication,
    is_repeat,
)
from mutalyzer.description import Description
from mutalyzer.description_model import (
    get_reference_id,
    get_selector_id,
    variants_to_description,
)
from mutalyzer.protein import get_protein_description, in_frame_description
from mutalyzer.reference import get_protein_selector_model
from mutalyzer.util import get_inserted_sequence, get_location_length
from mutalyzer_crossmapper import Coding
from pydantic import BaseModel, model_validator
from schema import And, Optional, Or, Schema
from typing_extensions import NewType

logger = logging.getLogger(__name__)

# Mutalyzer Variant dictionary
Variant_Dict = NewType("Variant_Dict", Mapping[str, Any])

CHROMOSOME_ASSEMBLY = {
    "NC_000001.11": "GRCh38",
    "NC_000002.12": "GRCh38",
    "NC_000003.12": "GRCh38",
    "NC_000004.12": "GRCh38",
    "NC_000005.10": "GRCh38",
    "NC_000006.12": "GRCh38",
    "NC_000007.14": "GRCh38",
    "NC_000008.11": "GRCh38",
    "NC_000009.12": "GRCh38",
    "NC_000010.11": "GRCh38",
    "NC_000011.10": "GRCh38",
    "NC_000012.12": "GRCh38",
    "NC_000013.11": "GRCh38",
    "NC_000014.9": "GRCh38",
    "NC_000015.10": "GRCh38",
    "NC_000016.10": "GRCh38",
    "NC_000017.11": "GRCh38",
    "NC_000018.10": "GRCh38",
    "NC_000019.10": "GRCh38",
    "NC_000020.11": "GRCh38",
    "NC_000021.9": "GRCh38",
    "NC_000022.11": "GRCh38",
    "NC_000023.11": "GRCh38",
    "NC_000024.10": "GRCh38",
    "NC_012920.1": "GRCh38",
}


def sequence_from_description(d: Description) -> str:
    """Return the sequence form a description"""
    _id = d.input_model["reference"]["id"]
    sequence: str = d.references[_id]["sequence"]["seq"]
    return sequence


class Variant:
    """Class to store delins variants"""

    # fmt: off
    # Schema for the location specification of the indel model
    location_schema = Schema(
        {
            "type": "range",
            "start": {
                "type": "point",
                "position": int,
            },
            "end": {
                "type": "point",
                "position": int,
            },
        }
    )

    # Schema for the inserted/deleted entries of the indel model
    inserted_deleted_schema = Schema(
        And( # Inserted must be 0 or 1 items
            lambda n: len(n) <= 1,
            [
                {
                    "sequence": Or(str, []),
                    "source": "description",
                    Optional("inverted") : True
                },
            ],
        ),
    )

    # Full schema for the indel model
    schema = Schema(
        {
            "type": "deletion_insertion",
            "source": "reference",
            "location": location_schema,
            Optional("inserted"): inserted_deleted_schema,
            Optional("deleted"): inserted_deleted_schema,
        }
    )
    # fmt: on

    def __init__(
        self,
        start: int,
        end: int,
        inserted: str = "",
        deleted: str = "",
    ):
        if start > end:
            raise ValueError(f"End ({end}) must be after start ({start})")
        self.start = start  # zero based
        self.end = end  # exclusive
        self.inserted = inserted

        if len(deleted) > 1:
            raise ValueError("deleted sequence is only defined for SNPS, not indels")
        self.deleted = deleted

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        start = self.start
        end = self.end
        inserted = self.inserted
        deleted = self.deleted
        return f"Variant({start=}, {end=}, inserted={inserted}, deleted={deleted})"

    def before(self, other: "Variant") -> bool:
        return self.end <= other.start

    def after(self, other: "Variant") -> bool:
        return self.start >= other.end

    def inside(self, other: "Variant") -> bool:
        return self.start >= other.start and self.end <= other.end

    def overlap(self, other: "Variant") -> bool:
        self_ends_in_other = self.end > other.start and self.end <= other.end
        self_starts_in_other = self.start >= other.start and self.start < other.end

        return any(
            [
                self_starts_in_other,
                self_ends_in_other,
                self.inside(other),
                other.inside(self),
            ]
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Variant):
            raise NotImplementedError

        return (
            self.start == other.start
            and self.end == other.end
            and self.inserted == other.inserted
            and self.deleted == other.deleted
        )

    def __lt__(self, other: "Variant") -> bool:
        if not isinstance(other, Variant):
            raise NotImplementedError
        if self.overlap(other):
            msg = f"Overlapping variants '{self}' and '{other}' cannot be sorted"
            raise ValueError(msg)

        return self.start < other.start

    @staticmethod
    def _validate_schema(model: Mapping[str, Any]) -> None:
        """Validate the structure of the mutalyzer delins model

        This can be very complex, and we only support the most common cases.
        """

        Variant.schema.validate(model)

    @staticmethod
    def _model_is_repeat(model: Mapping[str, Any]) -> bool:
        """Determine if model is a repeat"""
        # Determine if the model is a repeat
        repeat_schema = Schema(
            And(  # An empty list would be a deletion, not a repeat
                lambda n: len(n) == 1,
                [
                    {
                        "source": "description",
                        "sequence": str,
                        "repeat_number": {"type": "point", "value": int},
                        Optional("inverted"): True,
                    }
                ],
            )
        )

        inserted = model.get("inserted")
        is_repeat: bool = repeat_schema.is_valid(inserted)

        return is_repeat

    @staticmethod
    def _model_repeat_to_delins(model: Mapping[str, Any]) -> dict[str, Any]:
        """Convert a repeat model to a delins model"""
        new_model = {k: v for k, v in model.items()}

        # Determine the sequence and repeats
        old_sequence = model["inserted"][0]["sequence"]
        repeats = model["inserted"][0]["repeat_number"].get("value", 1)

        # Expand the new sequence
        new_sequence = old_sequence * repeats

        inserted = {"sequence": new_sequence, "source": "description"}

        # If the model is defined on the reverse strand
        inverted = model["inserted"][0].get("inverted", False)
        if inverted:
            inserted["inverted"] = True

        new_model["inserted"] = [inserted]
        return new_model

    @staticmethod
    def _model_is_duplication(model: Mapping[str, Any]) -> bool:
        """Determine if model is a duplication"""
        # Determine if the model is a repeat
        location_schema = Schema(
            {
                "type": "range",
                "start": {
                    "type": "point",
                    "position": int,
                },
                "end": {
                    "type": "point",
                    "position": int,
                },
            }
        )
        duplication_schema = Schema(
            And(  # An empty list would be a deletion, not a repeat
                lambda n: len(n) == 1,
                [
                    {
                        "source": "reference",
                        "location": location_schema,
                        "repeat_number": {"type": "point", "value": int},
                    }
                ],
            )
        )

        inserted = model.get("inserted")
        is_duplication: bool = duplication_schema.is_valid(inserted)

        return is_duplication

    @staticmethod
    def _model_duplication_to_delins(
        model: Mapping[str, Any], sequence: str
    ) -> dict[str, Any]:
        """Convert a duplication model to a delins model"""
        if not sequence:
            raise ValueError("Variant: specify sequence to handle duplications")
        new_model = {k: v for k, v in model.items()}

        # Determine the start and end on the sequence
        inserted = model["inserted"][0]
        start = inserted["location"]["start"]["position"]
        end = inserted["location"]["end"]["position"]
        repeats = inserted["repeat_number"].get("value", 1)

        # Expand the new sequence
        new_sequence = sequence[start:end]

        new_model["inserted"] = [
            {
                "sequence": new_sequence,
                "repeat_number": {"type": "point", "value": repeats},
                "source": "description",
            }
        ]
        return new_model

    @staticmethod
    def _model_is_inversion(model: Mapping[str, Any]) -> bool:
        """Determine if model is an inversion"""
        # Determine if the model is a repeat
        location_schema = Schema(
            {
                "type": "range",
                "start": {
                    "type": "point",
                    "position": int,
                },
                "end": {
                    "type": "point",
                    "position": int,
                },
            }
        )
        duplication_schema = Schema(
            And(  # An empty list would be a deletion, not a repeat
                lambda n: len(n) == 1,
                [
                    {
                        "source": "reference",
                        "location": location_schema,
                        "inverted": True,
                    }
                ],
            )
        )

        inserted = model.get("inserted")
        is_duplication: bool = duplication_schema.is_valid(inserted)

        return is_duplication

    @staticmethod
    def _reverse_complement(seq: str) -> str:
        complement = {"A": "T", "T": "A", "G": "C", "C": "G"}
        return "".join((complement[nt] for nt in reversed(seq)))

    @staticmethod
    def _model_inversion_to_delins(
        model: Mapping[str, Any], sequence: str
    ) -> dict[str, Any]:
        """Convert an inversion model to a delins model"""
        if not sequence:
            raise ValueError("Variant: specify sequence to handle inversions")
        new_model = {k: v for k, v in model.items()}

        # Determine the start and end on the sequence
        inserted = model["inserted"][0]
        start = inserted["location"]["start"]["position"]
        end = inserted["location"]["end"]["position"]

        # Expand the new sequence
        new_sequence = Variant._reverse_complement(sequence[start:end])

        new_model["inserted"] = [
            {
                "sequence": new_sequence,
                "source": "description",
            }
        ]
        return new_model

    @classmethod
    def from_model(cls, model: Mapping[str, Any], sequence: str = "") -> "Variant":
        if Variant._model_is_inversion(model):
            model = Variant._model_inversion_to_delins(model, sequence)
        if Variant._model_is_duplication(model):
            model = Variant._model_duplication_to_delins(model, sequence)
        # Determine if the model is a repeat
        if Variant._model_is_repeat(model):
            model = Variant._model_repeat_to_delins(model)

        # Validate the delins model
        cls._validate_schema(model)

        start = model["location"]["start"]["position"]
        end = model["location"]["end"]["position"]

        # Store if the inserted and deleted sequences were inverted
        ins_inverted: bool | None = None
        del_inverted: bool | None = None

        inserted = model.get("inserted", [])

        # Sanity check, complex variants should have been normalized into a
        # delins representation
        if len(inserted) == 0:
            inserted = ""
        elif len(inserted) > 1:
            raise NotImplementedError(
                "Multiple records in 'inserted' are not supported"
            )
        else:  # inserted contains 1 item, as is expected
            if "sequence" not in inserted[0]:
                raise NotImplementedError(
                    "Missing sequence in 'inserted' is not supported"
                )
            if "repeat_number" in inserted[0]:
                raise NotImplementedError("Repeats in 'inserted' are not supported")

            ins_inverted = inserted[0].get("inverted", False)
            inserted = inserted[0]["sequence"]

        deleted = model.get("deleted")
        if deleted is not None:
            try:
                del_inverted = deleted[0].get("inverted", False)
                deleted = deleted[0]["sequence"]
            except KeyError:
                raise NotImplementedError("Complex Variant not supported")

        # Make sure inserted and deleted are both or neither inversed
        if ins_inverted is not None and del_inverted is not None:
            if ins_inverted != del_inverted:
                msg = "strand difference between inserted and deleted sequences are not supported"
                raise NotImplementedError(msg)

        # If inserted or deleted are inverted, we need to reverse complement
        # the inserted/deleted nucleotides
        if inserted and ins_inverted:
            inserted = cls._reverse_complement(inserted)
        if deleted and del_inverted:
            deleted = cls._reverse_complement(deleted)

        return Variant(
            start=start,
            end=end,
            inserted=inserted if inserted else "",
            deleted=deleted if deleted else "",
        )

    @classmethod
    def from_dict(cls, dict: Mapping[str, Any]) -> "Variant":
        """Create a Variant object from a dict representation of a Variant"""
        return cls(**dict)

    def to_model(self) -> Mapping[str, Any]:
        """Convert Variant to mutalyzer delins model"""

        # Specification of the location
        # fmt: off
        location = {
            "type": "range",
            "start": {
                "type": "point",
                "position": self.start
            },
            "end": {
                "type": "point",
                "position": self.end
            }
        }

        # Specification of the inserted sequence
        inserted_obj: dict[str, Any] = {
            "sequence": self.inserted,
            "source": "description"
        }

        if self.inserted:
            inserted = [inserted_obj]
        else:
            inserted = []
        # fmt: on

        model = {
            "location": location,
            "type": "deletion_insertion",
            "source": "reference",
            "inserted": inserted,
        }

        deletion_object: dict[str, Any] = {
            "sequence": self.deleted,
            "source": "description",
        }
        if self.deleted:
            model["deleted"] = [deletion_object]

        return model

    def genomic_coordinates(self, d: Description) -> tuple[int, int]:
        """Return genomic coordinates for Variant"""
        offset = get_offset(d)

        if offset is None:
            raise RuntimeError("Missing ensembl offset")

        return self.start + offset, self.end + offset


@dataclasses.dataclass
class Therapy:
    """Class to store genetic therapies"""

    name: str
    hgvsc: str
    description: str
    variants: Sequence[Variant]
    figure: str | None = None
    hgvsr: str | None = None
    hgvsp: str | None = None

    @classmethod
    def from_dict(cls, dict: Mapping[str, Any]) -> "Therapy":
        """Create a Therapy object from a dict representation of a Therapy"""
        v = [Variant.from_dict(x) for x in dict["variants"]]
        return cls(
            name=dict["name"],
            hgvsc=dict["hgvsc"],
            description=dict["description"],
            variants=v,
            figure=dict.get("figure"),
            hgvsr=dict.get("hgvsr"),
            hgvsp=dict.get("hgvsp"),
        )


class HGVS(BaseModel):
    description: str

    model_config = {
        "json_schema_extra": {
            "examples": [
                {
                    "description": "ENST00000357033.9:c.6439_6614del",
                }
            ]
        }
    }

    @model_validator(mode="after")
    def hgvs_parser(self) -> "HGVS":
        """Parse the HGVS description with mutalyzer-hgvs-parser"""
        hgvs_error = (
            mutalyzer_hgvs_parser.exceptions.UnexpectedCharacter,
            mutalyzer_hgvs_parser.exceptions.UnexpectedEnd,
        )
        try:
            mutalyzer_hgvs_parser.to_model(self.description)
        except hgvs_error as e:
            raise ValueError(e)
        return self


def combine_variants_deletion(
    variants: Sequence[Variant], deletion: Variant
) -> Sequence[Variant]:
    """Combine variants and a deletion, any variants that are contained in
    the deletion are discarded

    The resulting list of variants is sorted
    """
    if deletion.inserted:
        raise ValueError(f"{Variant} is not a pure deletion")

    # Ensure the variants are sorted, and do not overlap
    sorted_variants = sorted(variants)

    combined = list()
    for i in range(len(sorted_variants)):
        variant = sorted_variants[i]
        if variant.before(deletion):
            combined.append(variant)
        elif variant.inside(deletion):
            # Discard the current variant
            continue
        elif variant.after(deletion):
            combined.append(deletion)
            combined += sorted_variants[i:]
            break
        else:
            msg = f"Deletion '{deletion}' partially overlaps '{variant}"
            raise ValueError(msg)
    else:
        combined.append(deletion)

    return combined


def de_to_hgvs(variants: Any, sequences: Any) -> Sequence[Variant_Dict]:
    """
    Convert the description extractor variants to an HGVS format (e.g., a
    deletion insertion of one nucleotide is converted to a substitution).

    MODIFIED from mutalyzer to not perform the 3' shift
    """
    if len(variants) == 1 and variants[0].get("type") == "equal":
        new_variant = deepcopy(variants[0])
        new_variant.pop("location")
        return [new_variant]

    new_variants = []
    for variant in variants:
        if variant.get("type") == "inversion":
            new_variants.append(deepcopy(variant))
        elif variant.get("type") == "deletion_insertion":
            inserted_sequence = get_inserted_sequence(variant, sequences)
            if len(inserted_sequence) == 0:
                new_variants.append(delins_to_del(variant))
            elif (
                get_location_length(variant["location"]) == len(inserted_sequence) == 1
            ):
                new_variants.append(delins_to_substitution(variant, sequences))
            elif is_repeat(variant, sequences):
                new_variants.append(delins_to_repeat(variant, sequences))
            elif is_duplication(variant, sequences):
                new_variants.append(delins_to_duplication(variant, sequences))
            elif get_start(variant["location"]) == get_end(variant["location"]):
                new_variants.append(delins_to_insertion(variant))
            else:
                new_variants.append(delins_to_delins(variant))

    return new_variants


def to_cdot_hgvs(d: Description, variants: Sequence[Variant]) -> str:
    """Convert a list of _Variants to hgvs representation"""
    delins_model = [v.to_model() for v in variants]

    # Invert the deleted sequence if the transcript is on the reverse strand
    if d.is_inverted():
        for delins in delins_model:
            if "deleted" in delins:
                _del = delins["deleted"][0]
                _del["sequence"] = Variant._reverse_complement(_del["sequence"])
                _del["inverted"] = True

    variant_models = de_to_hgvs(delins_model, d.get_sequences())

    ref_id = get_reference_id(d.corrected_model)
    selector_id = get_selector_id(d.corrected_model)

    selector_model = get_protein_selector_model(
        reference=d.references[ref_id]["annotations"], selector_id=selector_id
    )

    description_model = {
        "type": "description_dna",
        "reference": {"id": ref_id, "selector": {"id": ref_id}},
        "coordinate_system": "c",
        "variants": variant_models,
    }

    cdot_locations = to_hgvs_locations(
        description_model,
        d.references,
        selector_model=selector_model,
    )["variants"]

    hgvs: str = variants_to_description(cdot_locations)
    return hgvs


T = TypeVar("T")


def sliding_window(items: Sequence[T], size: int = 1) -> Sequence[Sequence[T]]:
    adj: list[Sequence[T]] = list()
    for i in range(len(items) - size + 1):
        adj.append([x for x in items[i : i + size]])
    return adj


def _exon_string(exon_numbers: Sequence[int]) -> str:
    """Format the exon names for a variable number of exons"""
    if len(exon_numbers) == 1:
        return f"exon {exon_numbers[0]}"
    elif len(exon_numbers) == 2:
        return f"exons {exon_numbers[0]} and {exon_numbers[1]}"
    else:
        t = ", ".join((str(x) for x in exon_numbers[:-1]))
        return f"exons {t} and {exon_numbers[-1]}"


def skip_adjacent_exons(d: Description, number_to_skip: int = 1) -> Sequence[Therapy]:
    """Skipp all possible adjacent exons the specified Description"""
    exon_skips: list[Therapy] = list()

    skippable_exons = get_exons(d, in_transcript_order=True)[1:-1]
    sequence = sequence_from_description(d)
    variants = [
        Variant.from_model(v, sequence=sequence) for v in d.delins_model["variants"]
    ]
    logger.debug(f"Input variants: {variants}")

    for i, exons in enumerate(sliding_window(skippable_exons, size=number_to_skip), 2):
        # Generate the string of exon numbers
        exons_description = _exon_string(range(i, i + number_to_skip))

        if d.is_inverted():
            # Start of the first exon to skip
            start = exons[-1][0]
            # End of the last exon to skip
            end = exons[0][-1]

        else:
            start = exons[0][0]
            end = exons[-1][-1]

        exon_skip = Variant(start, end)

        # Combine the existing variants with the exon skip
        try:
            combined = combine_variants_deletion(variants, exon_skip)
        except ValueError as e:
            if number_to_skip == 1:
                msg = f"Cannot skip exon {exons_description}: {e}"
            else:
                msg = f"Cannot skip exons {exons_description}: {e}"
            logger.warn(msg)
            continue
        logger.debug(f"Skip {exons_description}({exons=}): {exon_skip=} {combined=}")

        description = f"The annotations based on the supplied variants, in combination with skipping {exons_description}."
        # Convert to c. notation (user facing)
        name = f"Skip {exons_description}"
        selector = d.get_selector_id()
        cdot_variants = to_cdot_hgvs(d, combined)
        hgvsc = f"{selector}:c.{cdot_variants}"
        hgvsr = f"{selector}:r.{cdot_variants}"
        description = description
        t = Therapy(
            name=name,
            hgvsc=hgvsc,
            hgvsr=hgvsr,
            hgvsp=protein_prediction(d, combined)[0],
            description=description,
            variants=combined,
        )
        exon_skips.append(t)

    return exon_skips


def generate_therapies(d: Description) -> Sequence[Therapy]:
    """Wrapper around the different therapies"""
    therapies: list[Therapy] = list()
    # Skip a single exon
    therapies += skip_adjacent_exons(d, number_to_skip=1)
    # Skip two adjacent exons
    therapies += skip_adjacent_exons(d, number_to_skip=2)
    return therapies


def init_description(hgvs: str) -> Description:
    """
    Generate and initialize a Description for the specified HGVS

    Doesn't normalize the positions
    """
    d = Description(hgvs, stop_on_error=True)

    d.to_delins()
    d.de_hgvs_internal_indexing_model = d.delins_model
    d.construct_de_hgvs_internal_indexing_model()
    d.construct_de_hgvs_coordinates_model()
    d.construct_normalized_description()
    d.construct_protein_description()

    return d


def get_offset(d: Description) -> int:
    ref_id = get_reference_id(d.corrected_model)
    offset: int = (
        d.references.get(ref_id, {})
        .get("annotations", {})
        .get("qualifiers", {})
        .get("location_offset", 0)
    )
    return offset


def get_assembly_name(d: Description) -> str:
    """Extract the assembly name from a Description"""
    qualifiers = d.references["reference"]["annotations"]["qualifiers"]
    # ENS
    if "assembly_name" in qualifiers:
        return str(qualifiers["assembly_name"])

    # NC(NM)
    id = d.references["reference"]["annotations"]["id"]
    if id in CHROMOSOME_ASSEMBLY:
        return CHROMOSOME_ASSEMBLY[id]

    # NM
    raise ValueError(f"Unable to determine assembly for {d}")


def get_transcript_name(d: Description) -> str:
    """Extract the transcript name from a description"""
    name = d.get_selector_model()["id"]
    if name is None:
        raise ValueError(f"Unable to determine transcript name for {d}")
    else:
        return str(name)


def get_chrom_name(d: Description) -> str:
    # Try ensembl chromosome
    chrom = _get_ensembl_chrom_name(d)
    if chrom is not None:
        return str(chrom)

    # Try NCBI chromosome
    chrom = _get_ncbi_chrom_name(d)
    if chrom is not None:
        return str(chrom)

    raise RuntimeError(f"Unable to determine the chromosome for {d}")


def _get_ensembl_chrom_name(d: Description) -> str | None:
    ref_id = get_reference_id(d.corrected_model)
    chrom_name: str | None = (
        d.references.get(ref_id, {})
        .get("annotations", {})
        .get("qualifiers", {})
        .get("chromosome_number")
    )
    return chrom_name


def _get_ncbi_chrom_name(d: Description) -> str | None:
    chrom_name: str | None = d.input_model["reference"].get("id")
    return chrom_name


def get_strand(d: Description) -> str:
    # NM transcripts are on the forward strand by definition, and their strand
    # is not defined in the selector model from mutalyzer. Therefore, we set 1
    # as the default value.
    strand = d.get_selector_model()["location"].get("strand", 1)
    if strand == 1:
        return "+"
    elif strand == -1:
        return "-"
    else:
        raise ValueError(f"Unknown strand for Description object {d}")


def changed_protein_positions(
    reference: str, observed: str
) -> Sequence[tuple[int, int]]:
    """
    Extract the change protein positions (0 based)
    """
    deleted = list()
    for op in Levenshtein.opcodes(reference, observed):
        operation = op[0]
        ref_start = op[1]
        ref_end = op[2]

        if operation == "equal":
            continue
        elif operation == "insert":
            continue
        elif operation == "replace":
            deleted.append((ref_start, ref_end))
        elif operation == "delete":
            deleted.append((ref_start, ref_end))

    return deleted


def protein_prediction(
    d: Description, variants: Sequence[Variant]
) -> tuple[str, str, str]:
    """Call mutalyzer get_protein_description on a Description and list of Variants"""
    # Get required data structures from the Description
    ref_id = get_reference_id(d.corrected_model)
    selector_id = get_selector_id(d.corrected_model)
    selector_model = get_protein_selector_model(
        d.references[ref_id]["annotations"], selector_id=selector_id
    )

    # Convert the Variants to their delins model representation
    delins = [v.to_model() for v in variants]

    description, reference, predicted, *rest = get_protein_description(
        delins, d.references, selector_model
    )

    # If mutalyzer puts in an unknown protein prediction, we overwrite it
    if description.endswith(":p.?"):
        id = description.split(":p.?")[0]
        desc = in_frame_description(reference, predicted)[0]
        description = f"{id}:p.{desc}"

    return description, reference, predicted


def mutation_to_cds_effect(
    d: Description, variants: Sequence[Variant]
) -> list[tuple[int, int]]:
    """
    Determine the effect of the specified HGVS description on the CDS, on the genome

    Steps:
    - Use the protein prediction of mutalyzer to determine which protein
      residues are changed
    - Map this back to a deletion in c. positions to determine which protein
      annotations are no longer valid
    - Convert the c. positions to genome coordiinates as used by the UCSC
    NOTE that the genome range is similar to the UCSC annotations on the genome,
    i.e. 0 based, half open. Not to be confused with hgvs g. positions
    """
    # Determine the protein positions that were changed
    protein = protein_prediction(d, variants)
    reference, observed = protein[1], protein[2]

    # Keep track of changed positions on the genome
    changed_genomic = list()

    # Create crossmapper
    exons = d.get_selector_model()["exon"]
    cds = d.get_selector_model()["cds"]
    assert len(cds) == 1
    crossmap = Coding(exons, cds[0], inverted=d.is_inverted())

    for start, end in changed_protein_positions(reference, observed):
        # Calculate the nucleotide changed amino acids into a deletion in HGVS c. format

        # Internal coordinate positions
        start = crossmap.protein_to_coordinate(
            # nth amino acid, first nt of codon
            (start + 1, 1, 0, 0, 0)
        )
        end = (
            # nth amino acid, last nt of codon
            crossmap.protein_to_coordinate((end, 3, 0, 0, 0))
            + 1
        )

        if end < start:
            start, end = end - 1, start + 1

        v = Variant(start, end)

        changed_genomic.append(v.genomic_coordinates(d))

    return changed_genomic


def get_exons(
    description: Description, in_transcript_order: bool
) -> Sequence[tuple[int, int]]:
    """Get exons from a Mutalyzer Description object

    Positions are in the internal coordinate system
    """
    exons: Sequence[tuple[int, int]] = description.get_selector_model()["exon"]
    if in_transcript_order and description.is_inverted():
        return exons[::-1]

    return exons



import pytest

from gtgt.flask import organize_results, validate_user_input
from gtgt.mutalyzer import Therapy, Variant
from gtgt.transcript import Comparison, Result

INVALID_INPUTS = [
    # Non-HGVS input
    ("A", "Not a valid HGVS description"),
]


@pytest.mark.parametrize("input, summary", INVALID_INPUTS)
def test_invalid_inputs(input: str, summary: str) -> None:
    assert validate_user_input(input)["summary"] == summary


def test_organize_results() -> None:
    """
    Test grouping the results for rendering
    """
    # Input Result
    input_variant = Variant(10, 11, deleted="A", inserted="T")
    input_ = Result(
        therapy=Therapy(
            name="Input",
            hgvsc="",
            description="",
            variants=[input_variant],
        ),
        comparison=[],
    )

    wildtype = Result(
        therapy=Therapy(
            name="Wildtype",
            hgvsc="",
            description="",
            variants=[],
        ),
        comparison=[],
    )

    skip_variant_exon = Result(
        therapy=Therapy(
            name="Skip exon 3",
            hgvsc="",
            description="Skip exon with variant",
            variants=[Variant(5, 15, inserted="")],
        ),
        comparison=[],
    )

    skip_exon_before_variant = Result(
        therapy=Therapy(
            name="Skip exon 2",
            hgvsc="",
            description="Skip exon before variant",
            variants=[Variant(2, 5, inserted=""), input_variant],
        ),
        comparison=[],
    )

    skip_exon_after_variant = Result(
        therapy=Therapy(
            name="Skip exon 4",
            hgvsc="",
            description="Skip exon after the variant",
            variants=[input_variant, Variant(15, 25, inserted="")],
        ),
        comparison=[],
    )

    # List of results
    results = [
        wildtype,
        skip_variant_exon,
        input_,
        skip_exon_before_variant,
        skip_exon_after_variant,
    ]

    # Expected organized version of the results
    expected = {
        "input": [input_, wildtype],
        "modified": [skip_variant_exon],
        "all": [skip_exon_before_variant, skip_exon_after_variant],
    }

    assert organize_results(results) == expected



import pytest

from gtgt.variant_validator import (
    Links,
    Payload,
    extract_variant,
    parse_payload,
)


@pytest.fixture
def ids() -> Links:
    return Links(
        omim_ids=["1", "2"],
        gene_symbol="COL7A1",
        ensembl_gene_id="ENSG123",
        uniprot="Q123",
        decipher="3-4000000-C-G",
        genomic_variant="NC_123:c.100A>T",
        hgnc="HGNC:123",
        ucsc="uc123.3",
    )


FIELDS = [
    ("lovd", "https://databases.lovd.nl/shared/genes/COL7A1"),
    ("omim", [f"https://www.omim.org/entry/{id}" for id in [1, 2]]),
    (
        "gnomad",
        "https://gnomad.broadinstitute.org/variant/3-4000000-C-G?dataset=gnomad_r4",
    ),
    ("stringdb", "https://string-db.org/cgi/network?identifiers=COL7A1"),
]


@pytest.mark.parametrize("field, url", FIELDS)
def test_url(ids: Links, field: str, url: str) -> None:
    """
    GIVEN a Links object
    WHEN we query the url for a given field
    THEN we should get the correct URL
    """
    assert ids.url(field) == url


def test_url_dict(ids: Links) -> None:
    d = ids.url_dict()

    assert d["lovd"] == "https://databases.lovd.nl/shared/genes/COL7A1"
    assert d["omim_1"] == "https://www.omim.org/entry/1"
    assert d["omim_2"] == "https://www.omim.org/entry/2"


def test_parse_payload_unknown_flag() -> None:
    """
    GIVEN a payload from variant_validator with an unknown "flag"
    WHEN we parse the payload
    THEN we should raise a NotImplementedError
    """
    payload = {"flag": "weirdflag"}
    with pytest.raises(NotImplementedError, match="flag: weirdflag"):
        parse_payload(payload, variant="", assembly="HG38")


def test_parse_payload_warning() -> None:
    """
    GIVEN a payload from variant_validator with a warning flag
    WHEN we parse the payload
    THEN we should raise a ValueError with the text of the warning
    """
    payload = {
        "flag": "warning",
        "validation_warning_1": {
            "validation_warnings": [
                "InvalidFieldError: The transcript ENST00000296930.10 is not in the RefSeq data set. Please select Ensembl",
                "WeirdMadeUpError: I don't feel like looking up that variant",
            ],
        },
    }
    with pytest.raises(ValueError, match="WeirdMadeUpError"):
        parse_payload(payload, variant="", assembly="HG38")


def test_extract_variant_payload_variant_was_normalized() -> None:
    """
    GIVEN a payload from a non-normalized variant
    WHEN we receive the reply for the normalized variant
    THEN we should be able to parse the payload
    """
    payload = {"normalized_variant": {"submitted_variant": "not_normalized"}}

    assert extract_variant(payload, "not_normalized") == {
        "submitted_variant": "not_normalized"
    }


def test_extract_variant_missing_variant() -> None:
    """
    GIVEN a payload from VariantValidator which lacks the input variant
    WHEN we attempt to extract the input variant
    THEN we should get a ValueError
    """
    payload: Payload = dict()
    with pytest.raises(ValueError, match="100A>T"):
        extract_variant(payload, "100A>T")


def test_parse_valid_payload() -> None:
    """
    GIVEN a valid payload from variant_validator
    WHEN we parse the payload
    THEN we should get a reduced payload with the relevant fields
    """
    payload = {
        "flag": "gene_variant",
        "100A>T": {
            "submitted_variant": "100A>T",
            "gene_ids": {
                "omim_id": [
                    164040,
                ],
                "ensembl_gene_id": "ENSG00000181163",
                "ucsc_id": "uc032vtc.2",
            },
            "gene_symbol": "NPM1",
            "annotations": {"db_xref": {"hgnc": "HGNC:7910"}},
            "primary_assembly_loci": {
                "hg38": {
                    "vcf": {
                        "alt": "CTCTG",
                        "chr": "chr5",
                        "pos": "171410539",
                        "ref": "C",
                    },
                    "hgvs_genomic_description": "NC_000005.10:g.171410540_171410543dup",
                }
            },
        },
    }

    expected = {
        "omim_ids": [164040],
        "gene_symbol": "NPM1",
        "ensembl_gene_id": "ENSG00000181163",
        "hgnc": "HGNC:7910",
        "ucsc": "uc032vtc.2",
        "decipher": "5-171410539-C-CTCTG",
        "genomic_variant": "NC_000005.10:g.171410540_171410543dup",
    }

    reply = parse_payload(payload, variant="100A>T", assembly="hg38")

    assert reply == expected



from typing import Any

import pytest

from gtgt import Bed
from gtgt.bed import _range_to_size_start

# This is just a type alias
from gtgt.range import Range


@pytest.fixture
def bed() -> Bed:
    return Bed("chr1", 0, 11)


def test_bed_properties(bed: Bed) -> None:
    assert bed.chrom == "chr1"
    assert bed.chromStart == 0
    assert bed.chromEnd == 11


def test_default_values_simple_attributes(bed: Bed) -> None:
    assert bed.name == "."
    assert bed.score == 0
    assert bed.strand == "."


def test_default_values_display_attributes(bed: Bed) -> None:
    assert bed.thickStart == 0
    assert bed.thickEnd == 11
    assert bed.itemRgb == (0, 0, 0)


def test_defaul_values_blocks() -> None:
    bed = Bed("chr1", 5, 10)
    # Block positions are relative to chromStart, so the first block should
    # start at 0, not 5
    assert bed.blockStarts == [0]
    assert bed.blockSizes == [5]


def test_from_blocks_one() -> None:
    """
    Test making a Bed record from a single Range
    """
    bed = Bed.from_blocks("chr1", [(0, 10)])
    assert bed.chromStart == 0
    assert bed.chromEnd == 10


def test_from_blocks_multiple() -> None:
    """
    Test making a Bed record from a bunch of ranges

    0 1 2 3 4 5 6 7 8 9 10
        - -   -   - - - -
    """
    bed = Bed.from_blocks("chr1", [(2, 4), (5, 6), (7, 11)])
    assert bed.chromStart == 2
    assert bed.chromEnd == 11
    assert bed.blockStarts == [0, 3, 5]
    assert bed.blockSizes == [2, 1, 4]


def test_default_values_blocks(bed: Bed) -> None:
    assert bed.blockCount == 1
    assert bed.blockSizes == [11]
    assert bed.blockStarts == [0]


def test_blocks_interface(bed: Bed) -> None:
    for start, end in bed.blocks():
        assert start == 0
        assert end == 11


# Bed records, and their corresponding representation in Bed format
bed_records = [
    (
        # Default values
        Bed("chr1", 0, 11, ".", 0),
        "chr1	0	11	.	0	.	0	11	0,0,0	1	11	0",
    ),
    (
        Bed(
            chrom="chr1",
            chromStart=0,
            chromEnd=11,
            name="name",
            score=5,
            strand="+",
            thickStart=8,
            thickEnd=10,
            itemRgb=(42, 42, 42),
            blockCount=2,
            blockSizes=[3, 4],
            blockStarts=[0, 7],
        ),
        "chr1	0	11	name	5	+	8	10	42,42,42	2	3,4	0,7",
    ),
]


@pytest.mark.parametrize("bed, line", bed_records)
def test_bed_roundtrip(bed: Bed, line: str) -> None:
    """Test writing a bed to string format"""
    # Convert bed record to line
    from_bed = str(bed)
    # Convert line to Bed record
    from_line = Bed.from_bedfile(line)

    # Check that the line from Bed is as expected
    assert from_bed == line
    # Check that the Bed record from line is as expected
    assert from_line == bed


bedfile_lines = [
    # Minimum BED record, with only 3 columns
    ("chr1", 0, 10),
    # With name column
    ("chr1", 0, 10, "Name"),
    # With score column
    ("chr1", 0, 10, "Name", 9999),
    # With strand column
    ("chr1", 0, 10, "Name", 9999, "+"),
    # With thickStart column
    ("chr1", 0, 10, "Name", 9999, "strand", 0),
    # With thickEnd column
    ("chr1", 0, 10, "Name", 9999, "strand", 0, 10),
    # With itemRgb column
    ("chr1", 0, 10, "Name", 9999, "strand", 0, 10, "255,255,255"),
    # With blockCount column
    ("chr1", 0, 10, "Name", 9999, "strand", 0, 10, "255,255,255", 1),
    # With blockSizes(5,2) and blockStarts(0,8) columns
    ("chr1", 0, 10, "Name", 9999, "strand", 0, 10, "255,255,255", 2, "5,2", "0,8"),
]


@pytest.mark.parametrize("columns", bedfile_lines)
def test_from_bedfile(columns: tuple[Any, ...]) -> None:
    Bed.from_bedfile("\t".join(map(str, columns)))


range_start_size = [
    # Range, offset, size, start
    ((0, 10), 0, 10, 0),
    ((10, 20), 10, 10, 0),
    ((10, 20), 5, 10, 5),
]


@pytest.mark.parametrize("range_, offset, size, start", range_start_size)
def test_range_to_blocks(range_: Range, offset: int, size: int, start: int) -> None:
    assert _range_to_size_start(range_, offset) == (size, start)


# Things that cannot be used to intersect a Bed record
not_intersectable = [
    (1, NotImplementedError),
    # If the strand is different, throw a value error
    (Bed("chr1", 0, 10, strand="+"), ValueError),
]


@pytest.mark.parametrize("intersector, error", not_intersectable)
def test_non_intersectable_for_bed(intersector: Any, error: Any, bed: Bed) -> None:
    """Test that we raise an error"""
    with pytest.raises(error):
        bed.intersect(intersector)


# fmt: off
intersect_bed = [
    # Bed before, Bed to intersect with, Bed after
    # If the intersector is on another chromosome, there is no overlap
    (Bed("chr1", 0, 10), Bed("chr2", 0, 10), Bed("chr1", 0, 0)),
    # If the intersector is the same, we should get the same record back
    (Bed("chr1", 0, 10,), Bed("chr1", 0, 10), Bed("chr1", 0, 10)),
    # If the intersector does not overlap at all, we should get an empty record back
    (Bed("chr1", 5, 10), Bed("chr1", 20, 30), Bed("chr1", 5, 5)),
    # The intersector does not overlap, and the record has multiple blocks
    (
        Bed("chr1", 5, 10, blockCount=2, blockSizes=[2, 2], blockStarts=[0, 3]),
        Bed("chr1", 20, 30),
        Bed("chr1", 5, 5)
    ),
    # The intersector with multiple blocks does not overlap
    (
        Bed("chr1", 20, 30),
        Bed("chr1", 5, 10, blockCount=2, blockSizes=[2, 2], blockStarts=[0, 3]),
        Bed("chr1", 20, 20)
    ),
    # Both have multiple blocks, and do not overlap
    (
        Bed("chr1", 20, 30, blockCount=2, blockSizes=[5, 5], blockStarts=[0, 5]),
        Bed("chr1", 5, 10, blockCount=2, blockSizes=[2, 2], blockStarts=[0, 3]),
        Bed("chr1", 20, 20)
    ),
    # Both have multiple blocks, and the first block overlaps
    (
        Bed("chr1", 20, 30, blockCount=2, blockSizes=[5, 5], blockStarts=[0, 5]),
        Bed("chr1", 5, 25, blockCount=2, blockSizes=[2, 2], blockStarts=[0, 18]),
        Bed("chr1", 23, 25)
    ),
    # Both have multiple blocks, and each block partially overlaps
    #
    #    0 1 2 3 4 5 6 7 8 9
    # A  - - - -   - -     -
    # S  - -   - - -       - - - - - -
    # E  - -   -   -       -
    (
        Bed("chr1", 0, 10, blockSizes=[4, 2, 1], blockStarts=[0, 5, 9]),
        Bed("chr1", 0, 15, blockSizes=[2, 3, 6], blockStarts=[0, 3, 9]),
        Bed("chr1", 0, 10, blockSizes=[2, 1, 1, 1], blockStarts=[0, 3, 5, 9]),
    ),
]
# fmt: on


@pytest.mark.parametrize("before, intersector, after", intersect_bed)
def test_intersect_bed(before: Bed, intersector: Bed, after: Bed) -> None:
    new = before.intersect(intersector)
    assert new == after


def test_zero_bed_object() -> None:
    bed = Bed("chr1", 5, 10)
    bed._zero_out()
    assert bed.chromEnd == 5
    assert bed.thickStart == 5
    assert bed.thickEnd == 5
    assert bed.blockCount == 1
    assert bed.blockSizes == [0]
    assert bed.blockStarts == [0]


def test_update_bed_empty_ranges(bed: Bed) -> None:
    """
    Given an emtpy rangelist
    When we update a Bed record with the empty list
    The Bed record should be zeroed out
    """
    ranges: list[Range] = list()
    bed.update(ranges)

    # Test that the Bed record has been zeroed out
    assert bed.chromStart == 0
    assert bed.chromEnd == 0


def test_update_bed_with_ranges_thick() -> None:
    """
    If a record has thickStart or thickEnd set (not default)
    When we try to update the record with ranges
    Then we raise a NotImplementedError
    """
    record = Bed("chr1", 0, 10, thickStart=2, thickEnd=9)
    ranges: list[Range] = [(4, 8)]
    with pytest.raises(NotImplementedError):
        record.update(ranges)


def test_update_bed_with_ranges(bed: Bed) -> None:
    """Test updating a Bed record by providing a list of ranges"""
    ranges: list[Range] = [(1, 4), (5, 6), (7, 9)]
    bed.update(ranges)

    assert bed.chromStart == 1
    assert bed.chromEnd == 9
    assert bed.blockCount == 3
    assert bed.blockSizes == [3, 1, 2]
    assert bed.blockStarts == [0, 4, 6]


# fmt: off
invalid_bed = [
    # ThickStart before chromStart
    (
        ("chr1", 10, 20, ".", 0, "+", 1),
        "thickStart outside of record"
    ),
    # ThickStart after chromEnd
    (
        ("chr1", 10, 20, ".", 0, "+", 21),
        "thickStart outside of record"
    ),
    # ThickEnd before chromStart
    (
        ("chr1", 10, 20, ".", 0, "+", 10, 0),
        "thickEnd outside of record"
    ),
    # ThickEnd after chromEnd
    (
        ("chr1", 10, 20, ".", 0, "+", 10, 21),
        "thickEnd outside of record"
    ),
    # ThickEnd before thickStart
    (
        ("chr1", 10, 20, ".", 0, "+", 15, 10),
        "thickEnd before thickStart"
    ),
    # incorrect blockCount
    (
        ("chr1", 10, 20, ".", 0, "+", 10, 20, (0,0,0), 2),
        "blockCount(.*) does not match the number of blocks(.*)",
    ),
    # Mismatch in number of fields between blockSizes and blockStarts
    (
        ("chr1", 10, 20, ".", 0, "+", 10, 20, (0,0,0), 2, [2,3], [1]),
        "number of values differs between blockSizes(.*) and blockStarts(.*)",
    ),
    # Blocks must not overlap
    (
        ("chr1", 10, 20, ".", 0, "+", 10, 20, (0,0,0), 2, [2,3], [0,1]),
        "Blocks must not overlap",
    ),
    # Block extends over the end of the Bed region
    (
        ("chr1", 10, 20, ".", 0, "+", 10, 20, (0,0,0), 1, [11], [0]),
        "Last block(.*) must end at self.chromEnd",
    ),
    # The first block must start at chromStart(=0, since this field is relative)
    (
        ("chr1", 10, 20, ".", 0, "+", 10, 20, (0,0,0), 1, [1], [9]),
        "The first block must start at chromStart",
    ),
    # The last block must end at chromEnd
    (
        ("chr1", 10, 20, ".", 0, "+", 10, 20, (0,0,0), 1, [8], [0]),
        "Last block(.*) must end at self.chromEnd",
    ),
    # Blocks must be in ascending order
    (
        ("chr1", 10, 20, ".", 0, "+", 10, 20, (0,0,0), 2, [1,1], [19,1]),
        "Blocks must be in ascending order",
    ),
]
# fmt: on


@pytest.mark.parametrize("bed, msg", invalid_bed)
def test_value_error(bed: str, msg: str) -> None:
    with pytest.raises(ValueError, match=msg):
        Bed(*bed)


def test_overlap_invalid_records() -> None:
    """Overlap is not supported for Bed records on different strands"""
    before = Bed("chr1", 0, 10, strand="+")
    selector = Bed("chr1", 0, 10, strand=".")
    with pytest.raises(ValueError):
        before.overlap(selector)


# fmt: off
bed_overlap = [
    # before, selector, after
    # Selector on different chromosome
    (Bed('chr1', 0, 10), Bed('chr2', 0, 10), Bed('chr1', 0, 0)),
    # Selector is after the record
    (Bed('chr1', 0, 10), Bed('chr1', 10, 20), Bed('chr1', 0, 0)),
    # Selector overlaps the start of the record by 1 bp
    (Bed('chr1', 10, 20), Bed('chr1', 0, 11), Bed('chr1', 10, 20)),
    # Selector overlaps the end of the record by 1 bp
    (Bed('chr1', 0, 10), Bed('chr1', 9, 20), Bed('chr1', 0, 10)),
    # Both the Record and Selector consist of multiple blocks
    (
        # Record
        Bed('chr1', 0, 10, blockSizes=[4, 2, 1], blockStarts=[0, 5, 9]),
        # Selector overlaps every block in Record, the first block twice
        Bed('chr1', 0, 15, blockSizes=[2, 3, 6], blockStarts=[0, 3, 9]),
        # Unchanged
        Bed('chr1', 0, 10, blockSizes=[4, 2, 1], blockStarts=[0, 5, 9]),
    )
]
# fmt: on


@pytest.mark.parametrize("before, selector, after", bed_overlap)
def test_bed_overlap(before: Bed, selector: Bed, after: Bed) -> None:
    new = before.overlap(selector)
    assert new == after


bed_subtract = [
    # Both A and the selector consist of multiple ranges
    # The first region in A has partial overlap with the first two selector regions
    # The second selector region (partially) overlaps the first two regions in A
    #
    #    0 1 2 3 4 5 6 7 8 9
    # A  - - - -   - -     -
    # S  - -   - - -       - - - - - -
    # E      -       -
    (
        Bed("chr1", 0, 10, blockSizes=[4, 2, 1], blockStarts=[0, 5, 9]),
        Bed("chr1", 0, 15, blockSizes=[2, 3, 6], blockStarts=[0, 3, 9]),
        Bed("chr1", 2, 7, blockSizes=[1, 1], blockStarts=[0, 4]),
    ),
    # Subtracting a record on a different chromosome shouldn't change anything
    (
        Bed("chr1", 0, 10, blockSizes=[4, 2, 1], blockStarts=[0, 5, 9]),
        Bed("chr2", 0, 10, blockSizes=[4, 2, 1], blockStarts=[0, 5, 9]),
        Bed("chr1", 0, 10, blockSizes=[4, 2, 1], blockStarts=[0, 5, 9]),
    ),
]


@pytest.mark.parametrize("a, selector, expected", bed_subtract)
def test_subtract_bed(a: Bed, selector: Bed, expected: Bed) -> None:
    a.subtract(selector)
    assert a == expected


def test_bed_size() -> None:
    """Test the size of a Bed record

    The size is the sum of all blocks
    """
    bed = Bed.from_blocks("chr1", [(0, 5), (10, 15)])
    assert bed.size == 10


not_comparable = [
    # Different chromosomes
    (Bed("chr1", 0, 0), Bed("chr2", 0, 0)),
    # Different names
    (Bed("chr1", 0, 0, name="a"), Bed("chr1", 0, 0, name="b")),
    # Different strands
    (Bed("chr1", 0, 0, strand="+"), Bed("chr1", 0, 0)),
    # The selector is zero
    (Bed("chr1", 0, 10), Bed("chr1", 0, 0)),
]


@pytest.mark.parametrize("a, b", not_comparable)
def test_non_comparable_bed(a: Bed, b: Bed) -> None:
    """Test that we raise an error"""
    with pytest.raises(ValueError):
        a.compare(b)


compare = [
    # A, B, A/B
    (Bed("chr1", 5, 10), Bed("chr1", 0, 10), 0.5),
    # A consists of 2 blocks
    (Bed.from_blocks("chr1", [(0, 10), (15, 20)]), Bed("chr1", 0, 100), 0.15),
    # A > B
    (Bed("chr1", 0, 100), Bed("chr1", 0, 10), 10),
    # A has zero size
    (Bed("chr1", 0, 0), Bed("chr1", 0, 10), 0),
]


@pytest.mark.parametrize("a, b, expected", compare)
def test_compare_bed(a: Bed, b: Bed, expected: float) -> None:
    assert a.compare(b) == expected


csv = [
    ("5", [5]),
    # Test handling trailing commas
    ("5,", [5]),
    ("1,2,", [1, 2]),
]


@pytest.mark.parametrize("csv, numbers", csv)
def test_csv_to_int(csv: str, numbers: list[int]) -> None:
    assert Bed._csv_to_int(csv) == numbers


def test_empty_bed() -> None:
    b = Bed()

    assert b.chrom == ""
    assert b.chromStart == 0
    assert b.chromEnd == 0

    # Derived
    assert b.size == 0

    # Test that an empty Bed record is False
    assert not b



import json

import pytest

from gtgt import Bed
from gtgt.mutalyzer import Therapy, Variant, init_description
from gtgt.transcript import Comparison, Result, Transcript


@pytest.fixture
def exons() -> Bed:
    exons = [(0, 10), (20, 40), (50, 60), (70, 100)]
    bed = Bed.from_blocks("chr1", exons)
    bed.name = "Exons"

    return bed


@pytest.fixture
def coding_exons() -> Bed:
    coding_exons = [(23, 40), (50, 60), (70, 72)]
    bed = Bed.from_blocks("chr1", coding_exons)
    bed.name = "Coding exons"

    return bed


@pytest.fixture
def transcript(exons: Bed, coding_exons: Bed) -> Transcript:
    """
    Bed records that make up a transcript
    Each positions shown here is 10x
    (i) means inferred by the init method

                  0 1 2 3 4 5 6 7 8 9
    exons         -   - -   -   - - -
    coding_exons      - -   -   -
    """
    return Transcript(rna_features=[exons], protein_features=[coding_exons])


def test_transcript_init(transcript: Transcript) -> None:
    assert transcript.rna_features[0].name == "Exons"
    assert transcript.protein_features[0].name == "Coding exons"


def test_empty_transcript() -> None:
    """Test creating and working with an empty transcript"""
    t = Transcript(rna_features=[], protein_features=[])

    # Test if we can get the records
    assert t.records() == []

    # Test if intersect works
    t.intersect(Bed("chr1", 10, 20))

    # Test if subtraction works
    t.subtract(Bed("chr1", 10, 20))

    # Test if mutating the transcript works
    d = init_description("ENST00000375549.8:c.10del")
    t.mutate(d, variants=[])
    assert t


intersect_selectors = [
    # Selector spans all exons
    (
        Bed("chr1", 0, 100),
        Bed(
            "chr1",
            0,
            100,
            name="Exons",
            blockSizes=[10, 20, 10, 30],
            blockStarts=[0, 20, 50, 70],
        ),
    ),
    # Selector on a different chromosome
    (Bed("chr2", 0, 100), Bed("chr1", 0, 0)),
    # Selector intersect the first exon
    (Bed("chr1", 5, 15), Bed("chr1", 5, 10)),
    # Selector intersects the last base of the first exon,
    # and the first base of the second exon
    (Bed("chr1", 9, 21), Bed("chr1", 9, 21, blockSizes=[1, 1], blockStarts=[0, 11])),
]


def test_transcript_init_no_coding(exons: Bed) -> None:
    t = Transcript(rna_features=[exons], protein_features=[])
    assert not t.coding_exons


@pytest.mark.parametrize("selector, exons", intersect_selectors)
def test_intersect_transcript(
    selector: Bed, exons: Bed, transcript: Transcript
) -> None:
    """Test if intersecting the Transcript updates the exons"""
    transcript.intersect(selector)

    # Ensure the name matches, it's less typing to do that here
    exons.name = "Exons"
    assert transcript.exons == exons


overlap_selectors = [
    # Selector spans all exons
    (
        Bed("chr1", 0, 100),
        Bed(
            "chr1",
            0,
            100,
            name="Exons",
            blockSizes=[10, 20, 10, 30],
            blockStarts=[0, 20, 50, 70],
        ),
    ),
    # Selector on a different chromosome
    (Bed("chr2", 0, 100), Bed("chr1", 0, 0)),
    # Selector intersect the first exon
    (Bed("chr1", 5, 15), Bed("chr1", 0, 10)),
    # Selector intersects the last base of the first exon,
    # and the first base of the second exon
    (Bed("chr1", 9, 21), Bed("chr1", 0, 40, blockSizes=[10, 20], blockStarts=[0, 20])),
]


@pytest.mark.parametrize("selector, exons", overlap_selectors)
def test_overlap_transcript(selector: Bed, exons: Bed, transcript: Transcript) -> None:
    """Test if overlapping the Transcript updates the exons"""
    transcript.overlap(selector)

    # Ensure the name matches, it's less typing to do that here
    exons.name = "Exons"
    assert transcript.exons == exons


subtract_selectors = [
    # Selector spans all exons
    (Bed("chr1", 0, 100), Bed("chr1", 0, 0)),
    # Selector on a different chromosome
    (
        Bed("chr2", 0, 100),
        Bed(
            "chr1",
            0,
            100,
            name="Exons",
            blockSizes=[10, 20, 10, 30],
            blockStarts=[0, 20, 50, 70],
        ),
    ),
    # Selector intersect the first exon
    (
        Bed("chr1", 5, 15),
        Bed("chr1", 0, 100, blockSizes=[5, 20, 10, 30], blockStarts=[0, 20, 50, 70]),
    ),
    # Selector intersects the last base of the first exon,
    # and the first base of the second exon
    (
        Bed("chr1", 9, 21),
        Bed("chr1", 0, 100, blockSizes=[9, 19, 10, 30], blockStarts=[0, 21, 50, 70]),
    ),
]


@pytest.mark.parametrize("selector, exons", subtract_selectors)
def test_subtract_transcript(selector: Bed, exons: Bed, transcript: Transcript) -> None:
    """Test if subtracting the Transcript updates the exons"""
    transcript.subtract(selector)

    # Ensure the name matches, it's less typing to do that here
    exons.name = "Exons"
    assert transcript.exons == exons


def test_compare_transcripts(transcript: Transcript, coding_exons: Bed) -> None:
    exon_blocks = [
        (0, 10),
        # (20, 40),  # Missing the second exon
        (50, 60),
        (70, 100),
    ]
    exons = Bed.from_blocks("chr1", exon_blocks)
    exons.name = "Exons"

    coding_blocks = [
        # (23, 40),  # Missing the second exon
        (50, 60),
        (70, 72),
    ]
    coding_exons = Bed.from_blocks("chr1", coding_blocks)
    coding_exons.name = "Coding exons"

    smaller = Transcript(rna_features=[exons], protein_features=[coding_exons])

    cmp = smaller.compare(transcript)

    assert cmp[0].percentage == pytest.approx(0.71, abs=0.01)
    assert cmp[1].percentage == pytest.approx(0.41, abs=0.01)


def test_Result_init() -> None:
    t = Therapy("skip exon 5", "ENST123:c.49_73del", "Try to skip exon 5", list())
    c = Comparison("Coding exons", 0.5, "100/200")

    r = Result(therapy=t, comparison=[c])

    assert True


def test_Result_comparison() -> None:
    t1 = Therapy("skip exon 5", "ENST123:c.49_73del", "Try to skip exon 5", list())
    c1 = Comparison("Coding exons", 0.5, "100/200")
    r1 = Result(therapy=t1, comparison=[c1])

    t2 = Therapy("skip exon 6", "ENST123:c.49_73del", "Try to skip exon 5", list())
    c2 = Comparison("Coding exons", 0.2, "100/200")
    r2 = Result(therapy=t2, comparison=[c2])

    # Results in the wrong order
    results = [r2, r1]

    # Highest scoring Results should come first
    assert sorted(results, reverse=True) == [r1, r2]


MUTATE = [
    (
        "=",  # HGVS mutation (no change)
        ((0, 87), (984, 1101), (1994, 2139), (7932, 8922)),  # Expected exons
        ((35, 87), (984, 1101), (1994, 2139), (7932, 8098)),  # Expected coding_exons
    ),
    (
        # This 1bp deletion introduces a STOP codon
        "40del",  # position (74, 75) was deleted on the RNA
        ((0, 74), (75, 87), (984, 1101), (1994, 2139), (7932, 8922)),
        #           STOP codon is conserved
        ((35, 74), (8095, 8098)),
    ),
    (
        # An in frame deletion that introduces a STOP codon
        "101_106del",  # Position (1032, 1038) was deleted on the RNA
        ((0, 87), (984, 1032), (1038, 1101), (1994, 2139), (7932, 8922)),
        #                      STOP codon
        ((35, 87), (984, 1031), (8095, 8098)),
    ),
    (
        # Dele exon 2 (in frame)
        "53_169del",
        ((0, 87), (1994, 2139), (7932, 8922)),  # Expected exons
        # Exon 2 starts in frame 1, so the deleted positions derived from the
        # changed protein positions are slightly different: (986, 1995)
        # The nucleotieds (984, 986) are used with (1996, 1997) to form the
        # first non-deleted amino acid
        ((35, 87), (1996, 2139), (7932, 8098)),  # Expected coding_exons
    ),
]

Ranges = list[tuple[int, int]]


@pytest.mark.parametrize("variant, exon_blocks, coding_exon_blocks", MUTATE)
def test_mutate_forward(
    variant: str, exon_blocks: Ranges, coding_exon_blocks: Ranges
) -> None:
    # Features of SDHD
    transcript = "ENST00000375549.8"
    chrom = "chr11"
    offset = 112086872

    # Exons and coding exons of SDHD
    exons = Bed.from_blocks(
        chrom,
        [
            (112086872, 112086959),
            (112087856, 112087973),
            (112088866, 112089011),
            (112094804, 112095794),
        ],
    )
    exons.name = "Exons"

    coding_exons = Bed.from_blocks(
        chrom,
        [
            (112086907, 112086959),
            (112087856, 112087973),
            (112088866, 112089011),
            (112094804, 112094970),
        ],
    )
    coding_exons.name = "Coding exons"

    # Variant to test
    d = init_description(f"{transcript}:c.{variant}")
    v = [
        Variant.from_model(delins)
        for delins in d.de_hgvs_internal_indexing_model["variants"]
    ]

    # Add the offset to the expected exon blocks
    exon_blocks = [(start + offset, end + offset) for start, end in exon_blocks]
    coding_exon_blocks = [
        (start + offset, end + offset) for start, end in coding_exon_blocks
    ]

    SDHD = Transcript(rna_features=[exons], protein_features=[coding_exons])
    SDHD.mutate(d, v)

    assert SDHD.exons and SDHD.exons.blocks() == exon_blocks
    assert SDHD.coding_exons and SDHD.coding_exons.blocks() == coding_exon_blocks


MUTATE = [
    (
        # HGVS mutation (no change)
        "=",
        # Expected exon
        (
            (0, 1405),
            (4197, 4290),
            (4891, 4981),
            (8482, 8633),
            (12173, 12270),
            (28715, 28766),
            (29802, 29880),
            (40181, 40284),
            (40722, 40845),
            (46925, 47765),
        ),
        # Expected coding exons
        (
            (1283, 1405),
            (4197, 4290),
            (4891, 4981),
            (8482, 8633),
            (12173, 12270),
            (28715, 28766),
            (29802, 29880),
            (40181, 40284),
            (40722, 40845),
            (46925, 47586),
        ),
    ),
    (
        # SNP which introduces a STOP codon
        "100G>T",
        # Expected exons, 1 nt was changed (position 47846)
        (
            (0, 1405),
            (4197, 4290),
            (4891, 4981),
            (8482, 8633),
            (12173, 12270),
            (28715, 28766),
            (29802, 29880),
            (40181, 40284),
            (40722, 40845),
            (46925, 47486),
            (47487, 47765),
        ),
        # Expected coding exons
        # (47486,47487) is the first deleted protein sequence
        # (1283, 1286) is the STOP codon
        ((1283, 1286), (47487, 47586)),
    ),
    (
        # In fram deletion which introduces a STOP codon
        "134_139del",
        # Expected exon, positions (47447, 47452) were deleted
        (
            (0, 1405),
            (4197, 4290),
            (4891, 4981),
            (8482, 8633),
            (12173, 12270),
            (28715, 28766),
            (29802, 29880),
            (40181, 40284),
            (40722, 40845),
            (46925, 47447),
            (47453, 47765),
        ),
        # Expected coding exons,
        # 47454 is the last conserved nucleotide
        # (1283, 1286) is the STOP codon
        ((1283, 1286), (47454, 47586)),
    ),
    (
        # Delete exon 2 (in frame)
        "662_784del",
        # Expected exon (40282, 40845) deleted
        (
            (0, 1405),
            (4197, 4290),
            (4891, 4981),
            (8482, 8633),
            (12173, 12270),
            (28715, 28766),
            (29802, 29880),
            (40181, 40284),
            (46925, 47765),
        ),
        # Expected coding exons, (40772, 40840) deleted on the protein level
        # Note that exon 2 starts in frame 1, so when going from the protein
        # sequence, a small region in exon 3 has also changed
        (
            (1283, 1405),
            (4197, 4290),
            (4891, 4981),
            (8482, 8633),
            (12173, 12270),
            (28715, 28766),
            (29802, 29880),
            (40181, 40282),
            (46925, 47586),
        ),
    ),
]


@pytest.mark.parametrize("variant, exon_blocks, coding_exon_blocks", MUTATE)
def test_mutate_reverse(
    variant: str, exon_blocks: Ranges, coding_exon_blocks: Ranges
) -> None:
    # Features of SDHD
    transcript = "ENST00000452863.10"
    chrom = "chr11"
    offset = 32387774

    # Exons and coding exons of WT1
    exons = Bed.from_blocks(
        chrom,
        [
            (32387774, 32389179),
            (32391971, 32392064),
            (32392665, 32392755),
            (32396256, 32396407),
            (32399947, 32400044),
            (32416489, 32416540),
            (32417576, 32417654),
            (32427955, 32428058),
            (32428496, 32428619),
            (32434699, 32435539),
        ],
    )
    exons.name = "Exons"

    coding_exons = Bed.from_blocks(
        chrom,
        [
            (32389057, 32389179),
            (32391971, 32392064),
            (32392665, 32392755),
            (32396256, 32396407),
            (32399947, 32400044),
            (32416489, 32416540),
            (32417576, 32417654),
            (32427955, 32428058),
            (32428496, 32428619),
            (32434699, 32435360),
        ],
    )
    coding_exons.name = "Coding exons"

    # Variant to test
    d = init_description(f"{transcript}:c.{variant}")
    v = [
        Variant.from_model(delins)
        for delins in d.de_hgvs_internal_indexing_model["variants"]
    ]

    # Add the offset to the expected exon blocks
    exon_blocks = [(start + offset, end + offset) for start, end in exon_blocks]
    coding_exon_blocks = [
        (start + offset, end + offset) for start, end in coding_exon_blocks
    ]

    WT1 = Transcript(rna_features=[exons], protein_features=[coding_exons])
    WT1.mutate(d, v)

    assert WT1.exons and WT1.exons.blocks() == exon_blocks
    assert WT1.coding_exons and WT1.coding_exons.blocks() == coding_exon_blocks


def test_Comparison_from_dict() -> None:
    """Test creating a Comparison from a dict"""
    c = Comparison("Wildtype", percentage=100, basepairs="100/100")

    d = {"name": "Wildtype", "percentage": 100, "basepairs": "100/100"}

    assert Comparison.from_dict(d) == c


def test_Result_from_dict() -> None:
    """Test  creating a Result from a dict"""

    r = Result(
        therapy=Therapy(
            name="wildtype",
            hgvsc="ENST:c.=",
            description="Free text",
            variants=[Variant(10, 12, inserted="ATG")],
        ),
        comparison=[Comparison("Exons", percentage=100, basepairs="120/120")],
    )

    therapy = {
        "name": "wildtype",
        "hgvsc": "ENST:c.=",
        "description": "Free text",
        "variants": [{"start": 10, "end": 12, "inserted": "ATG", "deleted": ""}],
    }
    comparison = [{"name": "Exons", "percentage": 100, "basepairs": "120/120"}]

    d = {"therapy": therapy, "comparison": comparison}

    assert Result.from_dict(d) == r



from typing import Any, Sequence

import pytest
from mutalyzer.description import Description
from mutalyzer.description_model import get_reference_id
from mutalyzer.protein import get_protein_description
from mutalyzer.reference import get_protein_selector_model

from gtgt.mutalyzer import (
    Variant,
    combine_variants_deletion,
    init_description,
    mutation_to_cds_effect,
    sequence_from_description,
    to_cdot_hgvs,
)


def SDHD_description() -> Description:
    """SDHD, on the forward strand"""
    return init_description("ENST00000375549.8:c.=")


def WT1_description() -> Description:
    """WT1, on the reverse strand"""
    return init_description("ENST00000452863.10:c.=")


class TestVariant:
    """Test the functionality of the Variant class

    Tests concerning the integration with mutalyzer are excluded
    """

    def test_Variant_to_str(self) -> None:
        """Test converting a Variant to string"""
        v = Variant(10, 11, "ATG")
        assert str(v) == "Variant(start=10, end=11, inserted=ATG, deleted=)"

    def test_Variant_snp_to_str(self) -> None:
        """SNPS are special, since they contain the inserted sequence"""
        # 10A>T
        v = Variant(10, 11, "T", "A")
        assert str(v) == "Variant(start=10, end=11, inserted=T, deleted=A)"

    def test_Variant_to_model_positions(self) -> None:
        """Test converting a variant to model"""
        v = Variant(10, 11, "ATG")
        model = v.to_model()

        assert model["location"]["start"]["position"] == 10
        assert model["location"]["end"]["position"] == 11
        assert model["inserted"][0]["sequence"] == "ATG"

        # Deleted entry is missing for deletions
        assert "deleted" not in model

    def test_Variant_to_model_snp(self) -> None:
        """Test converting a snp Variant to model"""
        # 10 A>T
        v = Variant(10, 11, "T", "A")
        model = v.to_model()

        assert model["inserted"][0]["sequence"] == "T"
        assert model["deleted"][0]["sequence"] == "A"

    def test_Variant_deleted_snp_only(self) -> None:
        """Test that deleted is only defined for SNPs, not larger indels

        (To match the behaviour of Mutalyzer)
        """
        with pytest.raises(ValueError):
            # 10_12delinsGG
            Variant(10, 12, "GG", "AT")

    def test_Variant_no_inserted_sequence(self) -> None:
        """
        GIVEN a Variant with an empty inserted sequence
        WHEN we convert to a model
        THEN inserted should be an empty list
        """
        v = Variant(10, 11)
        model = v.to_model()

        assert model["inserted"] == []

    def test_Variant_end_after_start(self) -> None:
        """Raise an error when the end is after the start"""
        with pytest.raises(ValueError):
            Variant(11, 10)

    def test_Variant_equal(self) -> None:
        """Test that variants are equal to themselves"""
        v1 = Variant(1, 2, inserted="A")
        v2 = Variant(1, 2, inserted="A")

        assert v1 == v2
        assert v2 == v1

    ORDERING = [
        # Ends are touching
        (Variant(1, 3), Variant(3, 5)),
        # Gap between variants
        (Variant(0, 1), Variant(2, 4)),
    ]

    @pytest.mark.parametrize("a, b", ORDERING)
    def test_Variant_relative_positions(self, a: Variant, b: Variant) -> None:
        """Variant a is before variant b"""
        assert a.before(b)
        assert b.after(a)

    INSIDE = [
        # Variants are inside themselves
        (Variant(0, 3), Variant(0, 3)),
        # Smaller variant a is inside b
        (Variant(1, 3), Variant(0, 3)),
        (Variant(0, 2), Variant(0, 3)),
        (Variant(1, 2), Variant(0, 3)),
    ]

    @pytest.mark.parametrize("a, b", INSIDE)
    def test_Variant_a_inside_b(self, a: Variant, b: Variant) -> None:
        """Variant a is inside variant b"""
        assert a.inside(b)

    NOT_INSIDE = [
        # a starts outside of b
        (Variant(0, 3), Variant(1, 3)),
        # a ends outside of b
        (Variant(1, 4), Variant(1, 3)),
        # b is inside of a
        (Variant(0, 3), Variant(1, 2)),
        # a before b
        (Variant(0, 3), Variant(3, 5)),
        # a after b
        (Variant(3, 5), Variant(1, 3)),
    ]

    @pytest.mark.parametrize("a, b", NOT_INSIDE)
    def test_Variant_a_not_inside_b(self, a: Variant, b: Variant) -> None:
        """Variant a is not inside variant b"""
        assert not a.inside(b)

    OVERLAP = [
        # b ends inside a
        (Variant(2, 5), Variant(1, 3)),
        # b fully inside a
        (Variant(2, 5), Variant(3, 4)),
        # b fully inside a, ends at end
        (Variant(2, 5), Variant(3, 5)),
        # b ends inside a
        (Variant(2, 5), Variant(2, 4)),
        # b starts in a, extends after
        (Variant(2, 5), Variant(3, 6)),
        # b start before a
        (Variant(2, 5), Variant(1, 5)),
        # a is inside b
        (Variant(2, 5), Variant(1, 6)),
        # a is inside b
        (Variant(2, 5), Variant(2, 6)),
        # a equals b
        (Variant(2, 5), Variant(2, 5)),
    ]

    @pytest.mark.parametrize("a, b", OVERLAP)
    def test_Variant_a_overlaps_b(self, a: Variant, b: Variant) -> None:
        """Variant a and b overlap"""
        assert a.overlap(b)
        assert b.overlap(a)

    NO_OVERLAP = [
        # a is after b
        (Variant(2, 5), Variant(1, 2)),
        # a is before b
        (Variant(2, 5), Variant(5, 6)),
    ]

    @pytest.mark.parametrize("a, b", NO_OVERLAP)
    def test_Variant_a_does_not_overlap_b(self, a: Variant, b: Variant) -> None:
        """Variant a and b do not overlap"""
        assert not a.overlap(b)
        assert not b.overlap(a)

    def test_Variant_ordering(self) -> None:
        """Test sorting variants by start position"""
        v1 = Variant(10, 11)
        v2 = Variant(0, 1)
        assert sorted([v1, v2]) == [v2, v1]

    def test_Variant_ordering_error(self) -> None:
        """Overlapping variants cannot be sorted, and raise an error"""
        v1 = Variant(10, 11)
        v2 = Variant(10, 11)
        with pytest.raises(ValueError):
            sorted([v1, v2])

    def test_Variant_from_dict(self) -> None:
        """Test creating a Variant from a dict"""
        v = Variant(10, 22, inserted="ATCGG", deleted="")
        d = {"start": 10, "end": 22, "inserted": "ATCGG", "deleted": ""}

        assert Variant.from_dict(d) == v


class TestCombineVariantsDeletion:
    """Test combining multiple variants with a deletion

    This is used in the context of exon skipping, where variants that are fully
    inside the deletion (exon skip) are removed
    """

    def test_combine_variants_deletion_error(self) -> None:
        """Test that a value error is raised if the Variant is not a deletion"""
        indel = Variant(10, 15, inserted="ATC")
        with pytest.raises(ValueError):
            combine_variants_deletion(list(), indel)

    def test_combine_variants_deletion_empty(self) -> None:
        """If ther are no other variants, the results only contain the deletion"""
        variants: list[Variant] = list()
        deletion = Variant(0, 10)
        assert combine_variants_deletion(variants, deletion) == [deletion]

    # fmt: off
    COMBINE = [
        ( # The variants are out of order
            # Variants
            [Variant(5, 7), Variant(2, 4)],
            # Deletion
            Variant(10, 11),
            # Expected
            [Variant(2,4), Variant(5, 7), Variant(10, 11)],
        ),
        ( # The deletion is before both variants
            # Variants
            [Variant(2, 4), Variant(5, 7)],
            # Deletion
            Variant(0, 1),
            # Expected
            [Variant(0, 1), Variant(2,4), Variant(5, 7)],
        ),
        ( # The deletion is between the variants
            # Variants
            [Variant(2, 4), Variant(5, 7)],
            # Deletion
            Variant(4, 5),
            # Expected
            [Variant(2,4), Variant(4, 5), Variant(5, 7)],
        ),
        ( # The deletion is after both variants
            # Variants
            [Variant(2, 4), Variant(5, 7)],
            # Deletion
            Variant(10, 11),
            # Expected
            [Variant(2,4), Variant(5, 7), Variant(10, 11)],
        ),
        ( # The deletion contains the first variant
            # Variants
            [Variant(2, 4), Variant(5, 7)],
            # Deletion
            Variant(1, 5),
            # Expected
            [Variant(1, 5), Variant(5, 7)],
        ),
        ( # The deletion contains the second variant
            # Variants
            [Variant(2,4), Variant(5, 7)],
            # Deletion
            Variant(4, 11),
            # Expected
            [Variant(2,4), Variant(4, 11)],
        ),
        ( # The deletion contains both variants
            # Variants
            [Variant(2,4), Variant(5, 7)],
            # Deletion
            Variant(2, 7),
            # Expected
            [Variant(2, 7)],
        ),
    ]
    # fmt: on

    @pytest.mark.parametrize("variants, deletion, expected", COMBINE)
    def test_combine_variants_deletion(
        self,
        variants: Sequence[Variant],
        deletion: Variant,
        expected: Sequence[Variant],
    ) -> None:
        combined = combine_variants_deletion(variants, deletion)
        assert combined == expected

    def test_combine_variants_deletion_variants_overlap_eachother(self) -> None:
        """Test that we get a value error if the variants overlap"""
        variants = [Variant(0, 2), Variant(1, 3)]
        deletion = Variant(10, 11)
        with pytest.raises(ValueError):
            combine_variants_deletion(variants, deletion)

    def test_combine_variants_deletion_variants_partially_overlap_deletion(
        self,
    ) -> None:
        """
        Test that we get a value error if one the variants partially overlaps
        the deletion
        """
        variants = [Variant(2, 4)]
        deletion = Variant(3, 11)
        with pytest.raises(ValueError):
            combine_variants_deletion(variants, deletion)


class TestVariantMutalyzerIntegration:
    """Test the interaction between Variants and mutalyzer"""

    def test_Variant_from_model(self) -> None:
        """Test creating a variant from a mutalyzer delins model"""
        # fmt: off
        delins_model = {
            "type": "deletion_insertion",
            "source": "reference",
            "location": {
                "type": "range",
                "start": {
                    "type": "point",
                    "position": 0
                },
                "end": {
                    "type": "point",
                    "position": 2
                }
            },
            "inserted": [
                {
                    "sequence": "ATC",
                    "source": "description"
                }
            ]
        }
        # fmt: on
        assert Variant.from_model(delins_model) == Variant(0, 2, inserted="ATC")

    def test_Variant_from_model_deletion(self) -> None:
        """Test creating a variant from a mutalyzer delins model

        If nothing is inserted, the inserted sequence is an emtpy list, but should
        be made an empty string in the _Variant object
        """
        # fmt: off
        delins_model = {
            "type": "deletion_insertion",
            "source": "reference",
            "location": {
                "type": "range",
                "start": {
                    "type": "point",
                    "position": 0
                },
                "end": {
                    "type": "point",
                    "position": 2
                }
            },
            "inserted": [
                {
                    "sequence": [],
                    "source": "description"
                }
            ]
        }
        # fmt: on
        assert Variant.from_model(delins_model) == Variant(0, 2)


class TestVariantMutalyzerForward(object):
    """Test the interaction between Variants and mutalyzer for transcripts on the forward strand"""

    transcript = "ENST00000375549.8"

    # SDHD, forward strand
    def get_empty_transcript(self) -> Description:
        return init_description(f"{self.transcript}:c.=")

    def delins_from_description(self, d: Description) -> dict[str, Any]:
        """Return the internal delins model from a Description object"""
        delins_models = d.de_hgvs_internal_indexing_model["variants"]
        assert len(delins_models) == 1
        delins: dict[str, Any] = d.de_hgvs_internal_indexing_model["variants"][0]
        return delins

    def protein_from_description(self, d: Description) -> str:
        """Return the predicted protein sequence form a description"""
        protein_sequence: str = d.protein["predicted"]
        return protein_sequence

    def sequence_from_description(self, d: Description) -> str:
        """Return the sequence form a description"""
        return sequence_from_description(d)

    def protein_from_variant(self, v: Variant, d: Description) -> str:
        """Return the predicted protein sequence from a Variant"""

        ref_id = get_reference_id(d.corrected_model)

        selector_model = get_protein_selector_model(
            d.references[ref_id]["annotations"], ref_id
        )

        # Get the sequence (needed for complex variants)
        delins_models = [v.to_model()]

        predicted: str
        desc, ref, predicted, *rest = get_protein_description(
            delins_models, d.references, selector_model
        )
        return predicted

    # fmt: off
    MUTATIONS_VARIANT = [
        # HGVS, coordinates on the genome,
        # A simple missense that changes a single amino acids
        (
            "13T>A", # cdot notation of the variant, not used
            [Variant(start=47, end=48, inserted="A", deleted="T")], # Variants
            (112086919, 112086922) # Chromosomal location of the protein change
        ),
        # A stop mutation which destroys most of the protein
        (
            "9_10insTAG",
            [Variant(start=44, end=44, inserted="TAG")],
            (112086916, 112094967)
        ),
        # A frameshift that is restored by an insertion
        # Note this gives two separate, adjacent regions
        (
            "[9_10insA;20del]",
            [
                Variant(start=44, end=44, inserted="A"),
                Variant(start=54, end=55),
            ],
            [
                (112086919, 112086925),
                (112086925, 112086928),
            ]
        ),
        # A bigger deletion
        (
            "13_21del",
            [Variant(start=47, end=56)],
            (112086919, 112086928)
        ),
        # An SNP that creates a STOP codon
        (
            "14G>A",
            [Variant(start=48, end=49, inserted="A", deleted="G")],
            (112086919, 112094967)
        ),
    ]
    # fmt: on
    @pytest.mark.parametrize("cdot, variants, expected", MUTATIONS_VARIANT)
    def test_mutation_to_cds_effect(
        self, cdot: str, variants: Sequence[Variant], expected: tuple[int, int]
    ) -> None:
        """
        GIVEN a HGVS transcript description for a transcript on the reverse strand
        WHEN we determine the CDS effect
        THEN we should get genome coordinates
        """
        hgvs = "ENST00000375549.8:c.="
        d = init_description(hgvs)

        if not isinstance(expected, list):
            e = [expected]
        else:
            e = expected

        assert mutation_to_cds_effect(d, variants) == e

    # Variants where the delins model can be used to initialise Variant.from_model
    ROUND_TRIP_VARIANTS = [
        "10C>T",
        "10del",
        "10_11insA",
        "10_11delinsGG",
    ]

    @pytest.mark.parametrize("variant", ROUND_TRIP_VARIANTS)
    def test_Variant_hgvs_round_trip(
        self,
        variant: str,
    ) -> None:
        """
        For some variants, the conversion between the mutalyzer delins model
        and a Variant can be reversed, this is what we test here.

        Note that this only works for variants on the forward strand, on the
        reverse strand, only deletions can be converted round trip
        """
        d = init_description(f"{self.transcript}:c.{variant}")
        delins_model = d.delins_model["variants"][0]

        v = Variant.from_model(delins_model)
        assert v.to_model() == delins_model

    # Variant that are not simple delins in mutalyzer
    REWRITE_VARIANTS = [
        # Equivalent to delins 8_9delinsTTTT
        ("8_9T[4]", Variant(42, 44, inserted="TTTT")),
        # Equivalent to 10_10delinsCCC
        ("10C[3]", Variant(44, 45, inserted="CCC")),
        # Equivalent to 10_13delinsCTCTCTCT
        ("10_13CT[4]", Variant(44, 48, inserted="CTCTCTCT")),
        # Equivalent to 10_10delinsC
        ("10dup", Variant(44, 45, inserted="CC")),
        # Equivalent to 10_11delinsAG:
        ("10_11inv", Variant(44, 46, inserted="AG")),
    ]

    @pytest.mark.parametrize("variant_description, expected", REWRITE_VARIANTS)
    def test_delins_complex_Variant(
        self,
        variant_description: str,
        expected: Variant,
    ) -> None:
        """Convert a complex variant into a Variant

        Here, a complex variant is defined as a variant that is not represented
        as a simple delins model in Mutalyzer. When creating a
        Variant.from_model(), the variant representation from Mutalyzer is
        converted to an equivalent (but not equal) delins representation
        """
        d = init_description(f"{self.transcript}:c.{variant_description}")
        delins_model = d.delins_model["variants"][0]
        # Extract the sequence from the Description object
        _id = d.input_model["reference"]["id"]
        sequence = d.references[_id]["sequence"]["seq"]

        assert Variant.from_model(delins_model, sequence=sequence) == expected

    TO_HGVS = [
        # SNP
        (Variant(44, 45, "T", "C"), "10C>T"),
        # Deletion
        (Variant(44, 45), "10del"),
        # Insertion
        (Variant(45, 45, "A"), "10_11insA"),
        # Insertion/Deletion
        (Variant(44, 46, "GG"), "10_11delinsGG"),
    ]

    @pytest.mark.parametrize("variant, expected", TO_HGVS)
    def test_Variant_to_hgvs(self, variant: Variant, expected: str) -> None:
        d = SDHD_description()
        assert to_cdot_hgvs(d, [variant]) == expected

    NOT_SUPPORTED = [
        # Uncertain repeat size
        "8_9T[4_5]",
        # Uncertain repeat start
        "(6_8)_9T[4]",
        # Uncertain repeat end
        "8_(9_10)T[4]",
        # Uncertain start position
        "(9_15)insA",
        # Uncertain end position
        "8_(9_10)del"
        # Insertion of a range
        "9_10ins14_20",
    ]

    @pytest.mark.parametrize("variant", NOT_SUPPORTED)
    def test_variant_not_supported(self, variant: str) -> None:
        """Test that we throw a NotImplemented error for complex variants"""
        with pytest.raises(Exception):
            init_description(f"{self.transcript}:c.{variant}")

    VARIANTS = [
        # A SNP
        ("10C>T", Variant(44, 45, inserted="T", deleted="C")),
        # A deletion
        ("10del", Variant(44, 45, inserted="")),
        # An insertion
        ("10_11insA", Variant(45, 45, inserted="A")),
        # Delins version of 10C>T (Note that the deleted part is lost)
        ("10_10delinsT", Variant(44, 45, inserted="T")),
        # A duplication
        ("10dup", Variant(45, 45, inserted="C")),
        # The same duplication, with Variant as a delins (the deleted part is
        # implicit)
        ("10dup", Variant(44, 45, inserted="CC")),
        # A duplication
        ("10_11dup", Variant(44, 44, inserted="CT")),
        # The same duplication, where Variant deletes the first "CT", and then
        # inserts it twice
        ("10_11dup", Variant(44, 46, inserted="CTCT")),
        # Inversion, equivalent to 10C>G
        ("10_10inv", Variant(44, 45, inserted="G")),
        ("10C>G", Variant(44, 45, inserted="G")),
        # Inversion, equivalent to 10_11delinsAG
        ("10_11inv", Variant(44, 46, inserted="AG")),
        ("10_11delinsAG", Variant(44, 46, inserted="AG")),
        # Inversion, not symetrical
        ("18_20inv", Variant(52, 55, inserted="AGC")),
        ("18_20delinsAGC", Variant(52, 55, inserted="AGC")),
        # Small mononucleotide repeat:
        # ref: GGTT  CTCT
        # obs: GGTTTTCTCT
        # hgvs notation: 9_10insTT, 8_9T[4]
        ("8_9T[4]", Variant(44, 44, inserted="TT")),
        # Repeat of multiple distinct nucleotides
        # ref: GGTTCTCT    GGA
        # obs: GGTTCTCTCTCTGGA
        # hgvs notation: 9_10insCTCT, 10_13CT[4]
        ("10_13CT[4]", Variant(44, 44, inserted="CTCT")),
    ]

    @pytest.mark.parametrize("hgvs,variant", VARIANTS)
    def test_hgvs_Variant_equivalence_via_protein(
        self, hgvs: str, variant: Variant
    ) -> None:
        """Test hgvs and Variant equivalence by comparing the protein prediction

        The goal here is to verify that the model representation of the Variant
        is usable by mutalyzer in the same way as the original HGVS description
        """
        d = init_description(f"{self.transcript}:c.{hgvs}")
        variant_protein = self.protein_from_variant(
            variant, self.get_empty_transcript()
        )
        description_protein = self.protein_from_description(d)

        assert variant_protein == description_protein

    VARIANTS = [
        ("10C>T", Variant(44, 45, inserted="T", deleted="C")),
        ("10del", Variant(44, 45, inserted="")),
        ("10_11insA", Variant(45, 45, inserted="A")),
        ("10_10delinsT", Variant(44, 45, inserted="T")),
        ("10dup", Variant(44, 45, inserted="CC")),
        ("10_11dup", Variant(44, 46, inserted="CTCT")),
        ("10_10inv", Variant(44, 45, inserted="G")),
        ("10_11inv", Variant(44, 46, inserted="AG")),
        ("18_20inv", Variant(52, 55, inserted="AGC")),
        # Repeat is a delins of the existing repeat units, and insertion of the
        # new repeat
        ("8_9T[4]", Variant(42, 44, inserted="TTTT")),
        ("10_13CT[4]", Variant(44, 48, inserted="CTCTCTCT")),
    ]

    @pytest.mark.parametrize("hgvs, variant", VARIANTS)
    def test_hgvs_Variant_delins_model(self, hgvs: str, variant: Variant) -> None:
        """Test hgvs and Variant delins model equivalence directly

        The goal here is to verify that the Variant.from_model structure is
        always a simple delins model, even for duplications, repeats and
        inversions.
        """
        d = init_description(f"{self.transcript}:c.{hgvs}")
        delins_model = self.delins_from_description(d)

        seq = self.sequence_from_description(d)
        variant_model = Variant.from_model(delins_model, sequence=seq)
        assert variant_model == variant

    COORDINATES = [
        (Variant(47, 50), (112_086_919, 112_086_922)),
        (Variant(53, 56), (112_086_925, 112_086_928)),
        (Variant(44, 8095), (112_086_916, 112_094_967)),
        (Variant(47, 56), (112_086_919, 112_086_928)),
        (Variant(47, 8095), (112_086_919, 112_094_967)),
    ]

    @pytest.mark.parametrize("variant, genomic_coordinates", COORDINATES)
    def test_Variant_to_genomic_coordinates(
        self, variant: Variant, genomic_coordinates: tuple[int, int]
    ) -> None:
        """Test converting Variant coordinates to genomic coordinates"""
        d = init_description(f"{self.transcript}:c.=")
        assert variant.genomic_coordinates(d) == genomic_coordinates


class TestVariantMutalyzerReverse(TestVariantMutalyzerForward):
    """Test the interaction between Variants and mutalyzer for transcripts on the reverse strand"""

    # WT1, reverse strand
    transcript = "ENST00000452863.10"

    # fmt: off
    MUTATIONS_VARIANT = [
        # Variant, coordinates on the genome
        # A simple missense that changes a single amino acids
        (
            "13T>A",
            [Variant(start=47573, end=47574, inserted="T", deleted="A")],
            (32435345, 32435348)
        ),
        # A stop mutation which destroys most of the protein
        (
            "9_10insTAG",
            [Variant(start=47577, end=47577, inserted="CTA")],
            (32389060, 32435351)
        ),
        # # # A frameshift that is restored by an insertion
        (
            "[10del;20_21insA]",
            [
                Variant(start=47576, end=47577),
                Variant(start=47566, end=47566, inserted="T"),
            ],
            (32435339, 32435351)
        ),
        # # # A frameshift that is restored by a bigger insertion
        (
            "[10del;20_21insATCGAATATGGGG]",
            [
                Variant(start=47566, end=47566, inserted="CCCCATATTCGAT"),
                Variant(start=47576, end=47577),
            ],
            (32435339, 32435351)),
        # # # A bigger deletion
        (
            "11_19del",
            [Variant(start=47567, end=47576)],
             (32435342, 32435351)
        ),
        # # # An inframe deletion that creates a STOP codon
        (
            "87_89del",
            [Variant(start=47497, end=47500)],
            (32389060, 32435276)
        ),
    ]
    # fmt: on
    @pytest.mark.parametrize("cdot, variants, expected", MUTATIONS_VARIANT)
    def test_mutation_to_cds_effect(
        self, cdot: str, variants: Sequence[Variant], expected: tuple[int, int]
    ) -> None:
        """
        GIVEN a HGVS transcript description for a transcript on the reverse strand
        WHEN we determine the CDS effect
        THEN we should get genome coordinates
        """
        hgvs = "ENST00000452863.10:c.="
        d = init_description(hgvs)

        assert mutation_to_cds_effect(d, variants) == [expected]

    @pytest.mark.parametrize("variant", ["10del"])
    def test_Variant_hgvs_round_trip(self, variant: str) -> None:
        """
        See the equivalent test for forward transcripts. For transcripts on the
        reverse strand, a round-trip conversion can only be performed for deletions
        """
        d = init_description(f"{self.transcript}:c.{variant}")
        delins_model = d.delins_model["variants"][0]

        v = Variant.from_model(delins_model)
        assert v.to_model() == delins_model

    # Variant that are not simple delins in mutalyzer
    REWRITE_VARIANTS = [
        # Equivalent to 10_13delinsCTCTCTCT
        (
            "10_13CT[4]",
            Variant(47573, 47577, inserted="AGAGAGAG"),
        ),
        # Equivalent to 10_10delinsCC
        ("10dup", Variant(47576, 47577, inserted="GG")),
    ]

    @pytest.mark.parametrize("variant_description, expected", REWRITE_VARIANTS)
    def test_delins_complex_Variant(
        self,
        variant_description: str,
        expected: Variant,
    ) -> None:
        """Convert a complex variant into a Variant

        Here, a complex variant is defined as a variant that is not represented
        as a simple delins model in Mutalyzer. When creating a
        Variant.from_model(), the variant representation from Mutalyzer is
        converted to an equivalent (but not equal) delins representation

        Note that for transcripts on the reverse strand, only deletions are
        represented by a pure delins in Mutalyzer
        """
        d = init_description(f"{self.transcript}:c.{variant_description}")
        delins_model = d.delins_model["variants"][0]
        # Extract the sequence from the Description object
        _id = d.input_model["reference"]["id"]
        sequence = d.references[_id]["sequence"]["seq"]

        assert Variant.from_model(delins_model, sequence=sequence) == expected

    TO_HGVS = [
        # SNP
        (Variant(47573, 47574, inserted="T", deleted="A"), "13T>A"),
        # Deletion
        (Variant(start=47576, end=47577), "10del"),
        # Insertion
        (Variant(start=47577, end=47577, inserted="CTA"), "9_10insTAG"),
        # Insertion/deletion
        (Variant(47566, 47569, inserted="GCA"), "18_20delinsTGC"),
    ]

    @pytest.mark.parametrize("variant, expected", TO_HGVS)
    def test_Variant_to_hgvs(self, variant: Variant, expected: str) -> None:
        d = WT1_description()
        assert to_cdot_hgvs(d, [variant]) == expected

    NOT_SUPPORTED = [
        # Uncertain repeat size
        "8_9T[4_5]",
        # Uncertain repeat start
        "(6_8)_9T[4]",
        # Uncertain repeat end
        "8_(9_10)T[4]",
        # Uncertain start position
        "(9_15)insA",
        # Uncertain end position
        "8_(9_10)del"
        # Insertion of a range
        "9_10ins14_20",
    ]

    @pytest.mark.parametrize("variant", NOT_SUPPORTED)
    def test_variant_not_supported(self, variant: str) -> None:
        """Test that we throw a NotImplemented error for complex variants"""
        with pytest.raises(Exception):
            init_description(f"{self.transcript}:c.{variant}")

    # Note that on the Variant the nucleotides are (reverse?) complemented
    VARIANTS = [
        # A SNP,
        ("10C>T", Variant(47576, 47577, inserted="A", deleted="G")),
        # A deletion
        ("10del", Variant(47576, 47577, inserted="")),
        # An insertion
        ("10_11insA", Variant(47576, 47576, inserted="T")),
        # Delins version of 10C>T (Note that the deleted part is lost)
        ("10_10delinsT", Variant(47576, 47577, inserted="A")),
        # A duplication
        ("10dup", Variant(47576, 47576, inserted="G")),
        # The same duplication, with Variant as a delins (the deleted part is
        # implicit)
        ("10dup", Variant(47575, 47576, inserted="GG")),
        ("10_10delinsCC", Variant(47576, 47577, inserted="GG")),
        # A duplication, note that "AG" is the reverse complement of "CT"
        ("10_11dup", Variant(47575, 47575, inserted="AG")),
        ("11_12insCT", Variant(47575, 47575, inserted="AG")),
        ("12_13delinsCTCT", Variant(47575, 47577, inserted="AGAG")),
        ("12_13delinsCTCT", Variant(47573, 47575, inserted="AGAG")),
        # Inversion, equivalent to 10C>G
        ("10_10inv", Variant(47576, 47577, inserted="C")),
        ("10C>G", Variant(47576, 47577, inserted="C")),
        # Inversion, equivalent to 10_11delinsAG
        ("10_11inv", Variant(47575, 47577, inserted="CT")),
        ("10_11delinsAG", Variant(47575, 47577, inserted="CT")),
        # Inversion, not symetrical
        ("18_20inv", Variant(47566, 47569, inserted="GCA")),
        ("18_20delinsTGC", Variant(47566, 47569, inserted="GCA")),
        # Small mononucleotide repeat
        # 7_8T[4],  Variant representation = 7_8delinsTTTT
        ("7_8T[4]", Variant(47578, 47580, inserted="AAAA")),
        # Repeat of multiple distinct nucleotides
        # 10_13CT[4], Variant representation = 10_13delinsCTCTCTCT
        ("10_13CT[4]", Variant(47573, 47577, inserted="AGAGAGAG")),
    ]

    @pytest.mark.parametrize("hgvs,variant", VARIANTS)
    def test_hgvs_Variant_equivalence_via_protein(
        self, hgvs: str, variant: Variant
    ) -> None:
        """Test hgvs and Variant equivalence by comparing the protein prediction

        The goal here is to verify that the model representation of the Variant
        is usable by mutalyzer in the same way as the original HGVS description
        """
        d = init_description(f"{self.transcript}:c.{hgvs}")
        variant_protein = self.protein_from_variant(
            variant, self.get_empty_transcript()
        )
        description_protein = self.protein_from_description(d)

        assert variant_protein == description_protein

    VARIANTS = [
        ("10C>T", Variant(47576, 47577, inserted="A", deleted="G")),
        ("10del", Variant(47576, 47577, inserted="")),
        ("10_11insA", Variant(47576, 47576, inserted="T")),
        ("10_10delinsT", Variant(47576, 47577, inserted="A")),
        # Delete the original 'C', and insert CC (reverse complement)
        ("10dup", Variant(47576, 47577, inserted="GG")),
        ("10_10delinsCC", Variant(47576, 47577, inserted="GG")),
        # Delete the original "TC", and insert CTCT (reverse complement)
        ("10_11dup", Variant(47575, 47577, inserted="AGAG")),
        ("12_13delinsCTCT", Variant(47573, 47575, inserted="AGAG")),
        # Inversion, equivalent to 10C>G
        ("10_10inv", Variant(47576, 47577, inserted="C")),
        # Inversion, equivalent to 10_11delinsAG
        ("10_11inv", Variant(47575, 47577, inserted="CT")),
        # Inversion, not symetrical
        ("18_20inv", Variant(47566, 47569, inserted="GCA")),
        ("18_20delinsTGC", Variant(47566, 47569, inserted="GCA")),
        # Small mononucleotide repeat
        # 7_8T[4],  Variant representation = 7_8delinsTTTT
        ("7_8T[4]", Variant(47578, 47580, inserted="AAAA")),
        # Repeat of multiple distinct nucleotides
        # 10_13CT[4], Variant representation = 10_13delinsCTCTCTCT
        ("10_13CT[4]", Variant(47573, 47577, inserted="AGAGAGAG")),
    ]

    @pytest.mark.parametrize("hgvs, variant", VARIANTS)
    def test_hgvs_Variant_delins_model(self, hgvs: str, variant: Variant) -> None:
        """Test hgvs and Variant delins model equivalence directly

        The goal here is to verify that the Variant.from_model structure is
        always a simple delins model, even for duplications, repeats and
        inversions.
        """
        d = init_description(f"{self.transcript}:c.{hgvs}")
        delins_model = self.delins_from_description(d)

        seq = self.sequence_from_description(d)
        variant_model = Variant.from_model(delins_model, sequence=seq)
        assert variant_model == variant

    COORDINATES = [
        (Variant(47571, 47574), (32_435_345, 32_435_348)),
        (Variant(1286, 47577), (32_389_060, 32_435_351)),
        (Variant(47565, 47577), (32_435_339, 32_435_351)),
        (Variant(47565, 47577), (32_435_339, 32_435_351)),
        (Variant(47568, 47577), (32_435_342, 32_435_351)),
        (Variant(1286, 47502), (32_389_060, 32_435_276)),
    ]

    @pytest.mark.parametrize("variant, genomic_coordinates", COORDINATES)
    def test_Variant_to_genomic_coordinates(
        self, variant: Variant, genomic_coordinates: tuple[int, int]
    ) -> None:
        """Test converting Variant coordinates to genomic coordinates"""
        d = init_description(f"{self.transcript}:c.=")
        assert variant.genomic_coordinates(d) == genomic_coordinates



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

from glob import glob
from os.path import basename
from os.path import splitext
from pathlib import Path

from setuptools import find_packages
from setuptools import setup


this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name='gtgt',
    version='0.2.9',
    license='AGPL-3.0',
    description='Genetic Therapy Generator Toolkit',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Redmar van den Berg',
    author_email='RedmarvandenBerg@lumc.nl',
    url='https://github.com/DCRT-LUMC/GTGT',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        'Topic :: Utilities',
    ],
    project_urls={
        'Changelog': 'https://github.com/DCRT-LUMC/GTGT/blob/main/CHANGELOG.rst',
        'Issue Tracker': 'https://github.com/DCRT-LUMC/GTGT/issues',
    },
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires='>=3.10',
    install_requires=[
        "pydantic",
        "setuptools",
        "mutalyzer>=3.1.1",
        "mutalyzer_hgvs_parser",
        "lark==1.2.2",
        "exonviz>=0.2.16",
    ],
    extras_require={
        "webserver": ["flask"],
    },
    setup_requires=[
        'pytest-runner',
    ],
    entry_points={
        'console_scripts': [
            'gtgt=gtgt.cli:main',
        ]
    },
)



from typing import Any

import pytest

from gtgt import Bed
from gtgt.range import Range
from gtgt.ucsc import _blocks_overlap, _track_to_range, _tracks_to_bed

payload = dict[str, Any]


@pytest.mark.parametrize(
    "tracks, expected",
    [
        # A single track
        (
            [
                {
                    "chrom": "chr11",
                    "chromStart": 10,
                    "chromEnd": 20,
                    "name": "name",
                    "blockCount": 1,
                },
            ],
            [Bed("chr11", 10, 20, name="name")],
        ),
        (
            # Two tracks with the same name
            [
                {
                    "chrom": "chr11",
                    "chromStart": 10,
                    "chromEnd": 20,
                    "name": "name",
                    "blockCount": 1,
                },
                {
                    "chrom": "chr11",
                    "chromStart": 30,
                    "chromEnd": 40,
                    "name": "name",
                    "blockCount": 1,
                },
            ],
            [
                Bed(
                    "chr11",
                    10,
                    40,
                    name="name",
                    blockCount=2,
                    blockSizes=[10, 10],
                    blockStarts=[0, 20],
                )
            ],
        ),
        (
            # Two tracks with different names do not get merged
            [
                {
                    "chrom": "chr11",
                    "chromStart": 10,
                    "chromEnd": 20,
                    "name": "name",
                    "blockCount": 1,
                },
                {
                    "chrom": "chr11",
                    "chromStart": 30,
                    "chromEnd": 40,
                    "name": "NAME",
                    "blockCount": 1,
                },
            ],
            [
                Bed("chr11", 10, 20, name="name"),
                Bed("chr11", 30, 40, name="NAME"),
            ],
        ),
        (
            # Two tracks with the same name, one with multiple blocks
            [
                # [(5, 10), (15, 25)]
                {
                    "chrom": "chr11",
                    "chromStart": 5,
                    "chromEnd": 20,
                    "name": "name",
                    "blockCount": 2,
                    "chromStarts": "0,10",  # Relative to chromStart
                    "blockSizes": "5, 10",
                },
                # [(30, 40)]
                {
                    "chrom": "chr11",
                    "chromStart": 30,
                    "chromEnd": 40,
                    "name": "name",
                    "blockCount": 1,
                },
            ],
            [
                Bed(
                    "chr11",
                    5,
                    40,
                    name="name",
                    blockCount=3,
                    blockSizes=[5, 10, 10],
                    blockStarts=[0, 10, 25],
                )
            ],
        ),
    ],
)
def test_tracks_to_bed(tracks: list[payload], expected: list[Bed]) -> None:
    # Update constant values for the tracks
    constant = {
        "reserved": "100,100,0",
        "score": 1000,
        "strand": "+",
    }
    for track in tracks:
        track.update(constant)

    # Update the constant values for the BED records
    for bed in expected:
        bed.score = 1000
        bed.strand = "+"
        bed.itemRgb = (100, 100, 0)

    assert _tracks_to_bed(tracks) == expected


@pytest.mark.parametrize(
    "track, ranges",
    [
        (
            # A single block
            {
                "chromStart": 10,
                "chromEnd": 20,
                "blockCount": 1,
                "chromStarts": "0",
                "blockSizes": "10",
            },
            [(10, 20)],
        ),
        (
            # A track with two blocks
            {
                "chromStart": 10,
                "chromEnd": 40,
                "blockCount": 2,
                "chromStarts": "0,20",  # Relative to chromStart
                "blockSizes": "10, 10",
            },
            [(10, 20), (30, 40)],
        ),
        (
            # A track with two blocks, with trailing commas
            {
                "chromStart": 10,
                "chromEnd": 40,
                "blockCount": 2,
                "chromStarts": "0,20,",  # Relative to chromStart
                "blockSizes": "10, 10,",
            },
            [(10, 20), (30, 40)],
        ),
    ],
)
def test_track_to_range(track: payload, ranges: list[Range]) -> None:
    """Test converting a single track to a list of ranges"""
    assert _track_to_range(track) == ranges


@pytest.mark.parametrize(
    "blocks, expected",
    [
        ([], False),
        ([(0, 10)], False),
        ([(0, 10), (0, 10)], True),
        ([(0, 10), (10, 20)], False),
        ([(0, 10), (10, 20), (8, 12), (0, 10)], True),
    ],
)
def test_blocks_overlap(blocks: list[Range], expected: bool) -> None:
    assert _blocks_overlap(blocks) == expected



import json
from collections.abc import Sequence
from itertools import zip_longest

import pytest
from mutalyzer.description import Description

from gtgt.mutalyzer import (
    Therapy,
    Variant,
    _exon_string,
    changed_protein_positions,
    get_assembly_name,
    get_chrom_name,
    get_exons,
    get_transcript_name,
    init_description,
    protein_prediction,
    skip_adjacent_exons,
    sliding_window,
)
from gtgt.transcript import Transcript


def SDHD_description() -> Description:
    """SDHD, on the forward strand"""
    return init_description("ENST00000375549.8:c.=")


def WT1_description() -> Description:
    """WT1, on the reverse strand"""
    return init_description("ENST00000452863.10:c.=")


def WT_Transcript() -> Transcript:
    d = WT1_description()
    return Transcript.from_description(d)


def test_one_adjacent_exonskip_forward() -> None:
    d = SDHD_description()
    results = [
        "ENST00000375549.8:c.53_169del",
        "ENST00000375549.8:c.170_314del",
    ]
    for output, expected in zip_longest(skip_adjacent_exons(d), results):
        assert output.hgvsc == expected


def test_two_adjacent_exonskip_SDHD() -> None:
    d = SDHD_description()
    results = [
        "ENST00000375549.8:c.53_314del",
    ]
    for output, expected in zip_longest(
        skip_adjacent_exons(d, number_to_skip=2), results
    ):
        assert output.hgvsc == expected


def test_no_possible_exonskip_SDHD() -> None:
    """
    GIVEN a transcript with 4 exons (2 can be skipped)
    WHEN we try to skip 3 adjacent exons
    THEN we should get an empty list of therapies
    """
    d = SDHD_description()
    assert skip_adjacent_exons(d, number_to_skip=3) == list()


def test_one_adjacent_exonskip_WT1() -> None:
    d = WT1_description()
    results = [
        "ENST00000452863.10:c.662_784del",
        "ENST00000452863.10:c.785_887del",
        "ENST00000452863.10:c.888_965del",
        "ENST00000452863.10:c.966_1016del",
        "ENST00000452863.10:c.1017_1113del",
        "ENST00000452863.10:c.1114_1264del",
        "ENST00000452863.10:c.1265_1354del",
        "ENST00000452863.10:c.1355_1447del",
    ]
    # for output, expected in zip_longest(skip_adjacent_exons(d), results):
    for output, expected in zip_longest(skip_adjacent_exons(d), results):
        assert output.hgvsc == expected


def test_two_adjacent_exonskip_WT1() -> None:
    d = WT1_description()
    results = [
        "ENST00000452863.10:c.662_887del",
        "ENST00000452863.10:c.785_965del",
        "ENST00000452863.10:c.888_1016del",
        "ENST00000452863.10:c.966_1113del",
        "ENST00000452863.10:c.1017_1264del",
        "ENST00000452863.10:c.1114_1354del",
        "ENST00000452863.10:c.1265_1447del",
    ]
    for output, expected in zip_longest(skip_adjacent_exons(d, 2), results):
        assert output.hgvsc == expected


def test_sliding_window_size_one() -> None:
    s = "ABCDEF"
    # assert list(sliding_window(s, 1)) == [[x] for x in "A B C D E F".split()]
    assert list(sliding_window(s, 1)) == [["A"], ["B"], ["C"], ["D"], ["E"], ["F"]]


def test_sliding_window_size_2() -> None:
    s = "ABCDEF"
    assert list(sliding_window(s, 2)) == [
        ["A", "B"],
        ["B", "C"],
        ["C", "D"],
        ["D", "E"],
        ["E", "F"],
    ]


@pytest.mark.parametrize(
    "variant",
    "13T>A 970del 970_971insA 997_999delinsTAA 1000dup 10_11inv 994_996A[9]".split(),
)
def test_analyze_supported_variant_types(variant: str) -> None:
    WT = WT_Transcript()
    hgvs = f"ENST00000452863.10:c.{variant}"
    WT.analyze(hgvs)


def test_analyze_transcript() -> None:
    # In frame deletion that creates a STOP codon
    # variant = "ENST00000452863.10:c.87_89del"
    # Frameshift in small in-frame exon 5
    variant = "ENST00000452863.10:c.970del"

    WT = WT_Transcript()

    results = WT.analyze(variant)

    # Test the content of the 'wildtype' result
    wildtype = results[0]
    assert wildtype.therapy.name == "Wildtype"
    coding_exons = wildtype.comparison[1]
    assert coding_exons.name == "Coding exons"
    assert coding_exons.percentage == 1.0

    input = results[1]
    assert input.therapy.name == "Input"
    assert input.therapy.hgvsc == variant
    coding_exons = input.comparison[1]
    # basepairs are not a float, so easier to match than .percentage
    assert coding_exons.basepairs == "990/1569"


@pytest.mark.xfail
def test_analyze_transcript_r_coordinate(WT: Transcript) -> None:
    """Test analyzing a transcript using the r. coordinate system

    Note, this test should pass, but r. variant are currently not supported
    """
    # In frame deletion that creates a STOP codon
    variant = "ENST00000452863.10:r.970del"

    results = WT.analyze(variant)

    # Test the content of the 'wildtype' result
    wildtype = results[0]
    assert wildtype.therapy.name == "Wildtype"
    coding_exons = wildtype.comparison[1]
    assert coding_exons.name == "coding_exons"
    assert coding_exons.percentage == 1.0

    input = results[1]
    assert input.therapy.name == "Input"
    assert input.therapy.hgvsc == variant
    coding_exons = input.comparison[1]
    # basepairs are not a float, so easier to match than .percentage
    assert coding_exons.basepairs == "18845/46303"


PROTEIN_EXTRACTOR = [
    # No protein description
    ("", "", []),
    # No change
    ("A", "A", []),
    # A single change
    ("A", "T", [(0, 1)]),
    # A single change on the second position
    ("AAA", "ATA", [(1, 2)]),
    # Change in a repeat region
    ("AA", "A", [(1, 2)]),
    ("AAA", "A", [(1, 3)]),
    # A delins
    ("AAA", "ATC", [(1, 3)]),
    # An insertion, which we ignore
    ("AAA", "AATA", []),
    # A delins of TAG, which is equivalent to two insertions
    ("AAA", "ATAGAA", []),
    # A delins which is equivalent to a deletion
    ("AAA", "AGGGA", [(1, 2)]),
    # Multiple deletions
    ("AAA", "TAT", [(0, 1), (2, 3)]),
]


@pytest.mark.parametrize("reference, observed, expected", PROTEIN_EXTRACTOR)
def test_changed_protein_positions(
    reference: str, observed: str, expected: list[tuple[int, int]]
) -> None:
    """
    GIVEN a referene and observed sequence
    WHEN we extrat the protein changes
    THEN we should get 0 based positions of the deleted residues
    """
    assert changed_protein_positions(reference, observed) == expected


def test_get_exons_forward() -> None:
    """Text extracting exons from a Description object"""
    d = SDHD_description()
    expected = (0, 87)

    assert get_exons(d, in_transcript_order=True)[0] == expected
    assert get_exons(d, in_transcript_order=False)[0] == expected


def test_exons_forward() -> None:
    """Text extracting exons from a Description object"""
    d = WT1_description()
    expected_transcript_order = (46925, 47765)
    expected_genomic_order = (0, 1405)

    assert get_exons(d, in_transcript_order=True)[0] == expected_transcript_order
    assert get_exons(d, in_transcript_order=False)[0] == expected_genomic_order


EXON_DESCRIPTION = [
    ([2], "exon 2"),
    ([3, 5], "exons 3 and 5"),
    ([3, 4, 5, 6], "exons 3, 4, 5 and 6"),
]


@pytest.mark.parametrize("exons, expected", EXON_DESCRIPTION)
def test_exon_string(exons: Sequence[int], expected: str) -> None:
    assert _exon_string(exons) == expected


def test_Therapy_from_dict() -> None:
    """Test creating a Therapy from a dict"""
    variants = [Variant(10, 11, inserted="A", deleted="T")]
    therapy = Therapy(
        "Wildtype", hgvsc="ENST:c.=", description="free text", variants=variants
    )

    d = {
        "name": "Wildtype",
        "hgvsc": "ENST:c.=",
        "description": "free text",
        "variants": [{"start": 10, "end": 11, "inserted": "A", "deleted": "T"}],
    }

    assert Therapy.from_dict(d) == therapy


def test_protein_prediction_unknown() -> None:
    """Test overwriting unknown protein prediction

    Sometimes, mutalyzer will only generate :p.? as a protein prediction
    If that happens, we use in_frame_description overwrite the protein
    description
    """
    d = SDHD_description()
    # Variants that give rise to a :p.? prediction from mutalyzer
    variants = [Variant(start=1031, end=1032), Variant(start=1994, end=2139)]

    id = "ENST00000375549.8(ENSP00000364699)"
    p_variant = "Leu35_Leu159delinsPheArgThrAspLeuSerGlnAsnGlyValGluCysSerThrTyrThrCysHisArgAlaThrIleGlyProTrpThrSerCysTyr"
    assert protein_prediction(d, variants)[0] == f"{id}:p.{p_variant}"


def test_transcript_from_description_WT1() -> None:
    """Test creating a Transcript from a Mutalyzer Description"""
    d = WT1_description()
    t = Transcript.from_description(d)
    # Manually verified block starts for the exons of WT1
    assert t.exons and t.exons.blockStarts == [
        0,
        4197,
        4891,
        8482,
        12173,
        28715,
        29802,
        40181,
        40722,
        46925,
    ]
    # Manually verified block starts for the coding exons of WT1
    assert t.coding_exons and t.coding_exons.blockStarts == [
        0,
        2914,
        3608,
        7199,
        10890,
        27432,
        28519,
        38898,
        39439,
        45642,
    ]


def test_transcript_from_description_SDHD() -> None:
    """Test creating a Transcript from a Mutalyzer Description"""
    d = SDHD_description()
    t = Transcript.from_description(d)
    # Manually verified block starts for the exons of SDHD
    assert t.exons and t.exons.blockStarts == [0, 984, 1994, 7932]
    # Manually verified block starts for the coding exons of SDHD
    assert t.coding_exons and t.coding_exons.blockStarts == [0, 949, 1959, 7897]


def test_transcript_from_NC_NM_forward() -> None:
    """Test creating a forward Transcript from a Mutalyzer NC(NM) description"""
    d = init_description("NC_000011.10(NM_003002.4):c.=")
    t = Transcript.from_description(d)

    assert t.exons and t.exons.blocks() == [
        (5026, 5113),
        (6010, 6127),
        (7020, 7165),
        (12958, 13948),
    ]
    assert t.coding_exons and t.coding_exons.blocks() == [
        (5061, 5113),
        (6010, 6127),
        (7020, 7165),
        (12958, 13124),
    ]


def test_analyze_NC_NM_forward() -> None:
    """Test mutating an NC(NM) transcript on the forward strand

    NM_003002 is SDHD
    """
    hgvs = "NC_000011.10(NM_003002.4):c.102del"
    d = init_description(hgvs)
    t = Transcript.from_description(d)

    results = t.analyze(hgvs)

    # Look at the results for skipping exon 2
    skip2 = results[2]
    assert skip2.therapy.name == "Skip exon 2"
    assert skip2.therapy.hgvsc == "NM_003002.4:c.53_169del"
    assert skip2.therapy.hgvsp == "NC_000011.10(NP_002993.1):p.(Leu19_Ser57del)"

    # Check the remaining basepairs for the exons
    exons = skip2.comparison[0]
    assert exons.basepairs == "1222/1339"

    # Check the remaining basepairs for the coding exons
    coding_exons = skip2.comparison[1]
    assert coding_exons.basepairs == "361/480"


def test_analyze_NM_forward() -> None:
    """Test mutating an NM transcript on the forward strand

    NM_003002 is SDHD
    """
    hgvs = "NM_003002.4:c.102del"
    d = init_description(hgvs)
    t = Transcript.from_description(d)

    results = t.analyze(hgvs)

    # Look at the results for skipping exon 2
    skip2 = results[2]
    assert skip2.therapy.name == "Skip exon 2"
    assert skip2.therapy.hgvsc == "NM_003002.4:c.53_169del"
    assert skip2.therapy.hgvsp == "NM_003002.4(NP_002993.1):p.(Leu19_Ser57del)"

    # Check the remaining basepairs for the exons
    exons = skip2.comparison[0]
    assert exons.basepairs == "1222/1339"

    # Check the remaining basepairs for the coding exons
    coding_exons = skip2.comparison[1]
    assert coding_exons.basepairs == "361/480"


def test_transcript_from_NC_NM_reverse() -> None:
    """Test creating a reverse Transcript from a Mutalyzer NC(NM) description

    NM_012459.4 is TIMM8B
    """
    d = init_description("NC_000011.10(NM_012459.4):c.=")
    t = Transcript.from_description(d)

    assert t.exons and t.exons.blocks() == [(2953, 3616), (4793, 4910)]
    assert t.coding_exons and t.coding_exons.blocks() == [(3448, 3616), (4793, 4877)]


def test_analyze_NC_NM_reverse() -> None:
    """Test creating a reverse Transcript from a Mutalyzer NC(NM) description

    NM_012459.4 is TIMM8B
    """
    hgvs = "NC_000011.10(NM_012459.4):c.100del"
    d = init_description(hgvs)
    t = Transcript.from_description(d)

    results = t.analyze(hgvs)

    # TIMM8B only has 2 exons, so there will be no exon skips proposed
    assert len(results) == 2

    # Look at the results for the input variant
    input_results = results[1]
    assert input_results.therapy.name == "Input"
    assert (
        input_results.therapy.hgvsp == "NC_000011.10(NP_036591.3):p.(Glu34SerfsTer15)"
    )


def test_analyze_NM_reverse() -> None:
    """Test creating a reverse Transcript from a Mutalyzer NC(NM) description

    NM_012459.4 is TIMM8B
    """
    hgvs = "NM_012459.4:c.100del"
    d = init_description(hgvs)
    t = Transcript.from_description(d)

    results = t.analyze(hgvs)

    # TIMM8B only has 2 exons, so there will be no exon skips proposed
    assert len(results) == 2

    # Look at the results for the input variant
    input_results = results[1]
    assert input_results.therapy.name == "Input"
    assert input_results.therapy.hgvsp == "NM_012459.4(NP_036591.3):p.(Glu34SerfsTer15)"


@pytest.mark.parametrize(
    "transcript, chromosome",
    [
        ##  SDHD on chromosome 11  ##
        # ENST use chromosome numbers
        ("ENST00000375549.8", "11"),
        # NM transcripts are on their own 'chromosome'
        ("NM_003002.4", "NM_003002.4"),
        # NM on an NC should return the NC
        ("NC_000011.10(NM_003002.4)", "NC_000011.10"),
        ##  WT-1 on chromosome 11  ##
        # ENST use chromosome numbers
        ("ENST00000452863.10", "11"),
        ##  TIMM8B on chromosome  ##
        # NM transcripts are on their own 'chromosome'
        ("NM_012459.4", "NM_012459.4"),
        # NM on an NC should return the NC
        ("NC_000011.10(NM_012459.4)", "NC_000011.10"),
    ],
)
def test_chrom_name(transcript: str, chromosome: str) -> None:
    """Test extracting the chromosome name from a Description object"""
    d = init_description(f"{transcript}:c.=")
    assert get_chrom_name(d) == chromosome


@pytest.mark.parametrize(
    "transcript, assembly",
    [
        ##  SDHD on chromosome 11  ##
        # ENST use chromosome numbers
        ("ENST00000375549.8", "GRCh38"),
        # NM on an NC, we have a list of known chromsomes
        ("NC_000011.10(NM_003002.4)", "GRCh38"),
        ##  WT-1 on chromosome 11  ##
        ("ENST00000452863.10", "GRCh38"),
        ##  TIMM8B on chromosome  11 ##
        # NM on an NC should return the NC
        ("NC_000011.10(NM_012459.4)", "GRCh38"),
    ],
)
def test_assembly_name(transcript: str, assembly: str) -> None:
    """Test extracting the chromosome name from a Description object"""
    d = init_description(f"{transcript}:c.=")
    assert get_assembly_name(d) == assembly


def test_error_assembly_name_NM() -> None:
    d = init_description("NM_003002.4:c.=")
    with pytest.raises(ValueError):
        get_assembly_name(d)


@pytest.mark.parametrize(
    "transcript, expected",
    [
        ##  SDHD on chromosome 11  ##
        # ENST use chromosome numbers
        ("ENST00000375549.8", "ENST00000375549.8"),
        # NM on an NC, we have a list of known chromsomes
        ("NC_000011.10(NM_003002.4)", "NM_003002.4"),
        ##  WT-1 on chromosome 11  ##
        ("ENST00000452863.10", "ENST00000452863.10"),
        ##  TIMM8B on chromosome  11 ##
        # NM on an NC should return the NC
        ("NC_000011.10(NM_012459.4)", "NM_012459.4"),
    ],
)
def test_transcript_name(transcript: str, expected: str) -> None:
    """Test extracting the chromosome name from a Description object"""
    d = init_description(f"{transcript}:c.=")
    assert get_transcript_name(d) == expected
